"""
.. module:: theories.classy

:Synopsis: Managing the CLASS cosmological code
:Author: Jesus Torrado
         (import and ``get_cl`` based on MontePython's CLASS wrapper Benjamin Audren)

.. |br| raw:: html

   <br />

This module imports and manages the CLASS cosmological code.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   D. Blas, J. Lesgourgues, T. Tram,
   *The Cosmic Linear Anisotropy Solving System (CLASS). Part II: Approximation schemes*
   (`arXiv:1104.2933 <https://arxiv.org/abs/1104.2933>`_)

.. note::

   CLASS is renamed ``classy`` for most purposes within cobaya, due to CLASS's name being
   a python keyword.

Usage
-----

If you are using a likelihood that requires some observable from CLASS, simply add
``classy`` to the theory block.

You can specify any parameter that CLASS understands in the ``params`` block:

.. code-block:: yaml

   theory:
     classy:
       extra_args:
         [any param that CLASS understands]

   params:
       [any param that CLASS understands, fixed, sampled or derived]

If you want to use your own version of CLASS, you need to specify its location with a
``path`` option inside the ``classy`` block. If you do not specify a ``path``,
CLASS will be loaded from the automatic-install ``modules`` folder, if specified, or
otherwise imported as a globally-installed Python package. Cobaya will print at
initialisation where it is getting CLASS from.

.. _classy_modify:

Modifying CLASS
^^^^^^^^^^^^^^^

If you modify CLASS and add new variables, make sure that the variables you create are
exposed in the Python interface
(`instructions here <https://github.com/lesgourg/class_public/wiki/Python-wrapper>`__).
If you follow those instructions you do not need to make any additional modification in
Cobaya.

You can use the :doc:`model wrapper <cosmo_model>` to test your modification by
evaluating observables or getting derived quantities at known points in the parameter
space (set ``debug: True`` to get more detailed information of what exactly is passed to
CLASS).

In your CLASS modification, remember that you can raise a ``CosmoComputationError``
whenever the computation of any observable would fail, but you do not expect that
observable to be compatible with the data (e.g. at the fringes of the parameter
space). Whenever such an error is raised during sampling, the likelihood is assumed to be
zero, and the run is not interrupted.


Installation
------------

   .. _classy_install_warn:

.. warning::

   If the installation folder of CLASS is moved, due to CLASS hard-coding some folders,
   CLASS needs to be recompiled, either manually or by deleting the CLASS installation and
   repeating the ``cobaya-install`` command in the renamed *modules* folder.

   If you do not recompile CLASS, it causes a memory leak (`thanks to Stefan Heimersheim
   <https://github.com/CobayaSampler/cobaya/issues/10>`_).

Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CLASS, the easiest way to install it is using the
:doc:`automatic installation script <installation_cosmo>`. Just make sure that
``theory: classy:`` appears in one of the files passed as arguments to the installation
script.

Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CLASS or use an already modified version,
you should not use the automatic installation script. Use the method below instead.

CLASS's python interface utilizes the ``cython`` compiler. If typing ``cython`` in the
shell produces an error, install it with ``pip install cython --user``.

.. note::
   The fast way, assuming you are installing all your cosmological codes under
   ``/path/to/cosmo/``:

   .. code:: bash

      $ cd /path/to/cosmo/
      $ git clone https://github.com/lesgourg/class_public.git
      $ mv class_public CLASS
      $ cd CLASS
      $ make

   If the **second** line produces an error (because you don't have ``git`` installed),
   try

   .. code:: bash

      $ cd /path/to/cosmo/
      $ wget https://github.com/lesgourg/class_public/archive/master.zip
      $ unzip master.zip
      $ rm master.zip
      $ mv class_public-master CLASS
      $ cd CLASS
      $ make

If the instructions above failed, follow those in the
`official CLASS web page <http://class-code.net/>`_ to get CLASS compiled with the Python
interface ready.
"""
# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import sys
import os
import numpy as np
from copy import deepcopy
import logging
from collections import namedtuple, OrderedDict as odict
from time import time
from numbers import Number

# Local
from cobaya.theories._cosmo import _cosmo, PowerSpectrumInterpolator
from cobaya.log import HandledException
from cobaya.install import download_github_release, pip_install
from cobaya.conventions import _c_km_s, _T_CMB_K

# Result collector
collector = namedtuple("collector",
                       ["method", "args", "args_names", "kwargs", "arg_array", "post"])
collector.__new__.__defaults__ = (None, [], [], {}, None, None)

# default non linear code
non_linear_default_code = "halofit"


class classy(_cosmo):

    def initialize(self):
        """Importing CLASS from the correct path, if given, and if not, globally."""
        # If path not given, try using general path to modules
        if not self.path and self.path_install:
            self.path = os.path.join(
                self.path_install, "code", classy_repo_rename)
        if self.path:
            self.log.info("Importing *local* classy from " + self.path)
            classy_build_path = os.path.join(self.path, "python", "build")
            post = next(d for d in os.listdir(classy_build_path) if d.startswith("lib."))
            classy_build_path = os.path.join(classy_build_path, post)
            if not os.path.exists(classy_build_path):
                # If path was given as an install path, try to install global one anyway
                if self.path_install:
                    self.log.info("Importing *global* CLASS (because not installed).")
                else:
                    self.log.error("Either CLASS is not in the given folder, "
                                   "'%s', or you have not compiled it.", self.path)
                    raise HandledException
            else:
                # Inserting the previously found path into the list of import folders
                sys.path.insert(0, classy_build_path)
        else:
            self.log.info("Importing *global* CLASS.")
        try:
            from classy import Class, CosmoSevereError, CosmoComputationError
        except ImportError:
            self.log.error(
                "Couldn't find the CLASS python interface. "
                "Make sure that you have compiled it, and that you either\n"
                " (a) specify a path (you didn't) or\n"
                " (b) install the Python interface globally with\n"
                "     '/path/to/class/python/python setup.py install --user'")
            raise HandledException
        self.classy = Class()
        # Propagate errors up
        global CosmoComputationError, CosmoSevereError
        # Generate states, to avoid recomputing
        self.n_states = 3
        self.states = [{"params": None, "derived": None, "derived_extra": None,
                        "last": 0} for i in range(self.n_states)]
        # Dict of named tuples to collect requirements and computation methods
        self.collectors = {}
        # Additional input parameters to pass to CLASS
        self.extra_args = self.extra_args or {}
        # Add general CLASS stuff
        self.extra_args["output"] = self.extra_args.get("output", "")
        if "sBBN file" in self.extra_args:
            self.extra_args["sBBN file"] = (
                self.extra_args["sBBN file"].format(classy=self.path))
        # Set aliases
        self.planck_to_classy = self.renames
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []

    def current_state(self):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        return self.states[lasts.index(max(lasts))]

    def needs(self, **requirements):
        # Computed quantities required by the likelihood
        super(classy, self).needs(**requirements)
        for k, v in self._needs.items():
            # Products and other computations
            if k == "Cl":
                if any([("t" in cl.lower()) for cl in v]):
                    self.extra_args["output"] += " tCl"
                if any([(("e" in cl.lower()) or ("b" in cl.lower())) for cl in v]):
                    self.extra_args["output"] += " pCl"
                # For modern experiments, always lensed Cl's!
                self.extra_args["output"] += " lCl"
                self.extra_args["lensing"] = "yes"
                # For l_max_scalars, remember previous entries.
                self.extra_args["l_max_scalars"] = max(v.values())
                self.collectors[k] = collector(
                    method="lensed_cl", kwargs={"lmax": self.extra_args["l_max_scalars"]})
            elif k == "H":
                self.collectors[k] = collector(
                    method="Hubble",
                    args=[np.atleast_1d(v["z"])],
                    args_names=["z"],
                    arg_array=0)
                self.H_units_conv_factor = {"1/Mpc": 1, "km/s/Mpc": _c_km_s}
            elif k == "angular_diameter_distance":
                self.collectors[k] = collector(
                    method="angular_distance",
                    args=[np.atleast_1d(v["z"])],
                    args_names=["z"],
                    arg_array=0)
            elif k == "comoving_radial_distance":
                self.collectors[k] = collector(
                    method="z_of_r",
                    args_names=["z"],
                    args=[np.atleast_1d(v["z"])])
            elif k == "Pk_interpolator":
                self.extra_args["output"] += " mPk"
                self.extra_args["P_k_max_h/Mpc"] = max(
                    v.pop("k_max"), self.extra_args.get("P_k_max_h/Mpc", 0))
                self.add_z_for_matter_power(v.pop("z"))
                # Use halofit by default if non-linear requested but no code specified
                if v.get("nonlinear", False) and "non linear" not in self.extra_args:
                    self.extra_args["non linear"] = non_linear_default_code
                for pair in v.pop("vars_pairs", [["delta_tot", "delta_tot"]]):
                    if any([x != "delta_tot" for x in pair]):
                        self.log.error("NotImplemented in CLASS: %r", pair)
                        raise HandledException
                    self._Pk_interpolator_kwargs = {
                        "logk": True, "extrap_kmax": v.pop("extrap_kmax", None)}
                    name = "Pk_interpolator_%s_%s" % (pair[0], pair[1])
                    self.collectors[name] = collector(
                        method="get_pk_and_k_and_z",
                        kwargs=v,
                        post=(lambda P, k, z: PowerSpectrumInterpolator(
                            z, k, P.T, **self._Pk_interpolator_kwargs)))
            elif v is None:
                k_translated = self.translate_param(k, force=True)
                if k_translated not in self.derived_extra:
                    self.derived_extra += [k_translated]
            else:
                self.log.error("Requested product not known: %r", {k: v})
                raise HandledException
        # Derived parameters (if some need some additional computations)
        if any([("sigma8" in s) for s in self.output_params or requirements]):
            self.extra_args["output"] += " mPk"
            self.extra_args["P_k_max_h/Mpc"] = (
                max(1, self.extra_args.get("P_k_max_h/Mpc", 0)))
        # Adding tensor modes if requested
        if self.extra_args.get("r") or "r" in self.input_params:
            self.extra_args["modes"] = "s,t"
        # If B spectrum with l>50, or lensing, recommend using Halofit
        cls = self._needs.get("Cl", {})
        if (((any([("b" in cl.lower()) for cl in cls]) and
              max([cls[cl] for cl in cls if "b" in cl.lower()]) > 50) or
             any([("p" in cl.lower()) for cl in cls]) and
             not self.extra_args.get("non linear"))):
            self.log.warning("Requesting BB for ell>50 or lensing Cl's: "
                             "using a non-linear code is recommended (and you are not "
                             "using any). To activate it, set "
                             "'non_linear: halofit|hmcode|...' in classy's 'extra_args'.")
        # Cleanup of products string
        self.extra_args["output"] = " ".join(set(self.extra_args["output"].split()))
        # If no output requested, remove arguments that produce an error
        # (e.g. complaints if halofit requested but no Cl's computed.)
        # Needed for facilitating post-processing
        if not self.extra_args["output"]:
            for k in ["non linear"]:
                if k in self.extra_args:
                    self.log.info("Ignoring {%s: %r}, since no products requested.",
                                  k, self.extra_args[k])
                    self.extra_args.pop(k)
        # Finally, check that there are no repeated parameters between input and extra
        if set(self.input_params).intersection(set(self.extra_args)):
            self.log.error(
                "The following parameters appear both as input parameters and as CLASS "
                "extra arguments: %s. Please, remove one of the definitions of each.",
                list(set(self.input_params).intersection(set(self.extra_args))))
            raise HandledException

    def add_z_for_matter_power(self, z):
        if not hasattr(self, "z_for_matter_power"):
            self.z_for_matter_power = np.empty((0))
        self.z_for_matter_power = np.flip(np.sort(np.unique(np.concatenate(
            [self.z_for_matter_power, np.atleast_1d(z)]))), axis=0)
        self.extra_args["z_pk"] = " ".join(["%g" % zi for zi in self.z_for_matter_power])

    def translate_param(self, p, force=False):
        # "force=True" is used when communicating with likelihoods, which speak "planck"
        if self.use_planck_names or force:
            return self.planck_to_classy.get(p, p)
        return p

    def set(self, params_values_dict, i_state):
        # Store them, to use them later to identify the state
        self.states[i_state]["params"] = deepcopy(params_values_dict)
        # Prepare parameters to be passed: this-iteration + extra
        args = {self.translate_param(p): v for p, v in params_values_dict.items()}
        args.update(self.extra_args)
        # Generate and save
        self.log.debug("Setting parameters: %r", args)
        self.classy.struct_cleanup()
        self.classy.set(**args)

    def compute(self, _derived=None, cached=True, **params_values_dict):
        lasts = [self.states[i]["last"] for i in range(self.n_states)]
        try:
            if not cached:
                raise StopIteration
            # are the parameter values there already?
            i_state = next(i for i in range(self.n_states)
                           if self.states[i]["params"] == params_values_dict)
            # has any new product been requested?
            for product in self.collectors:
                next(k for k in self.states[i_state] if k == product)
            reused_state = True
            # Get (pre-computed) derived parameters
            if _derived == {}:
                _derived.update(self.states[i_state]["derived"])
            self.log.debug("Re-using computed results (state %d)", i_state)
        except StopIteration:
            reused_state = False
            # update the (first) oldest one and compute
            i_state = lasts.index(min(lasts))
            self.log.debug("Computing (state %d)", i_state)
            if self.timing:
                a = time()
            # Set parameters
            self.set(params_values_dict, i_state)
            # Compute!
            try:
                self.classy.compute()
            # "Valid" failure of CLASS: parameters too extreme -> log and report
            except CosmoComputationError:
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on.")
                return 0
            # CLASS not correctly initialized, or input parameters not correct
            except CosmoSevereError:
                self.log.error("Serious error setting parameters or computing results. "
                               "The parameters passed were %r and %r. "
                               "See original CLASS's error traceback below.\n",
                               self.states[i_state]["params"], self.extra_args)
                raise  # No HandledException, so that CLASS traceback gets printed
            # Gather products
            for product, collector in self.collectors.items():
                # Special case: sigma8 needs H0, which cannot be known beforehand:
                if "sigma8" in self.collectors:
                    self.collectors["sigma8"].args[0] = 8 / self.classy.h()
                method = getattr(self.classy, collector.method)
                arg_array = self.collectors[product].arg_array
                if arg_array is None:
                    self.states[i_state][product] = method(
                        *self.collectors[product].args, **self.collectors[product].kwargs)
                elif isinstance(arg_array, Number):
                    self.states[i_state][product] = np.zeros(
                        len(self.collectors[product].args[arg_array]))
                    for i, v in enumerate(self.collectors[product].args[arg_array]):
                        args = (list(self.collectors[product].args[:arg_array]) + [v] +
                                list(self.collectors[product].args[arg_array + 1:]))
                        self.states[i_state][product][i] = method(
                            *args, **self.collectors[product].kwargs)
                elif arg_array in self.collectors[product].kwargs:
                    value = np.atleast_1d(self.collectors[product].kwargs[arg_array])
                    self.states[i_state][product] = np.zeros(value.shape)
                    for i, v in enumerate(value):
                        kwargs = deepcopy(self.collectors[product].kwargs)
                        kwargs[arg_array] = v
                        self.states[i_state][product][i] = method(
                            *self.collectors[product].args, **kwargs)
                if collector.post:
                    self.states[i_state][product] = collector.post(
                        *self.states[i_state][product])
            # Prepare derived parameters
            d, d_extra = self._get_derived_all(derived_requested=(_derived == {}))
            if _derived == {}:
                _derived.update(d)
            self.states[i_state]["derived"] = odict(
                [[p, (_derived or {}).get(p)] for p in self.output_params])
            # Prepare necessary extra derived parameters
            self.states[i_state]["derived_extra"] = deepcopy(d_extra)
            if self.timing:
                self.n += 1
                self.time_avg = (time() - a + self.time_avg * (self.n - 1)) / self.n
                self.log.debug("Average running time: %g seconds", self.time_avg)
        # make this one the current one by decreasing the antiquity of the rest
        for i in range(self.n_states):
            self.states[i]["last"] -= max(lasts)
        self.states[i_state]["last"] = 1
        return 1 if reused_state else 2

    def _get_derived_all(self, derived_requested=True):
        """
        Returns a dictionary of derived parameters with their values,
        using the *current* state (i.e. it should only be called from
        the ``compute`` method).

        Parameter names are returned in CLASS nomenclature.

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        # Put all pamaremters in CLASS nomenclature (self.derived_extra already is)
        requested = [self.translate_param(p) for p in (
            self.output_params if derived_requested else [])]
        requested_and_extra = {
            p: None for p in set(requested).union(set(self.derived_extra))}
        # Parameters with their own getters
        if "rs_drag" in requested_and_extra:
            requested_and_extra["rs_drag"] = self.classy.rs_drag()
        elif "Omega_nu" in requested_and_extra:
            requested_and_extra["Omega_nu"] = self.classy.Omega_nu
        # Get the rest using the general derived param getter
        # No need for error control: classy.get_current_derived_parameters is passed
        # every derived parameter not excluded before, and cause an error, indicating
        # which parameters are not recognized
        requested_and_extra.update(
            self.classy.get_current_derived_parameters(
                [p for p, v in requested_and_extra.items() if v is None]))
        # Separate the parameters before returning
        # Remember: self.output_params is in sampler nomenclature,
        # but self.derived_extra is in CLASS
        derived = {
            p: requested_and_extra[self.translate_param(p)] for p in self.output_params}
        derived_extra = {p: requested_and_extra[p] for p in self.derived_extra}
        return derived, derived_extra

    def get_param(self, p):
        current_state = self.current_state()
        for pool in ["params", "derived", "derived_extra"]:
            value = deepcopy(
                current_state[pool].get(self.translate_param(p, force=True), None))
            if value is not None:
                return value
        self.log.error("Parameter not known: '%s'", p)
        raise HandledException

    def get_cl(self, ell_factor=False, units="muK2"):
        current_state = self.current_state()
        try:
            cls = deepcopy(current_state["Cl"])
        except:
            self.log.error(
                "No Cl's were computed. Are you sure that you have requested them?")
            raise HandledException
        # unit conversion and ell_factor
        ell_factor = ((cls["ell"] + 1) * cls["ell"] / (2 * np.pi))[2:] if ell_factor else 1
        units_factors = {"1": 1,
                         "muK2": _T_CMB_K * 1.e6,
                         "K2": _T_CMB_K}
        try:
            units_factor = units_factors[units]
        except KeyError:
            self.log.error("Units '%s' not recognized. Use one of %s.",
                           units, list(units_factors))
            raise HandledException
        for cl in cls:
            if cl not in ['pp', 'ell']:
                cls[cl][2:] *= units_factor ** 2 * ell_factor
        if "pp" in cls and ell_factor is not 1:
            cls['pp'][2:] *= ell_factor ** 2 * (2 * np.pi)
        return cls

    def _get_z_dependent(self, quantity, z):
        try:
            z_name = next(k for k in ["redshifts", "z"]
                          if k in self.collectors[quantity].kwargs)
            computed_redshifts = self.collectors[quantity].kwargs[z_name]
        except StopIteration:
            computed_redshifts = self.collectors[quantity].args[
                self.collectors[quantity].args_names.index("z")]
        i_kwarg_z = np.concatenate(
            [np.where(computed_redshifts == zi)[0] for zi in np.atleast_1d(z)])
        values = np.array(deepcopy(self.current_state()[quantity]))
        if quantity == "comoving_radial_distance":
            values = values[0]
        return values[i_kwarg_z]

    def get_H(self, z, units="km/s/Mpc"):
        try:
            return self._get_z_dependent("H", z) * self.H_units_conv_factor[units]
        except KeyError:
            self.log.error("Units not known for H: '%s'. Try instead one of %r.",
                           units, list(self.H_units_conv_factor))
            raise HandledException

    def get_angular_diameter_distance(self, z):
        return self._get_z_dependent("angular_diameter_distance", z)

    def get_comoving_radial_distance(self, z):
        return self._get_z_dependent("comoving_radial_distance", z)

    def get_Pk_interpolator(self):
        current_state = self.current_state()
        prefix = "Pk_interpolator_"
        return {k[len(prefix):]: deepcopy(v)
                for k, v in current_state.items() if k.startswith(prefix)}

    def close(self):
        self.classy.struct_cleanup()


# Installation routines ##################################################################

# Name of the Class repo/folder and version to download
classy_repo_name = "lesgourg/class_public"
classy_repo_rename = "classy"
classy_repo_version = "v2.7.2"


def is_installed(**kwargs):
    if not kwargs["code"]:
        return True
    return os.path.isfile(os.path.realpath(
        os.path.join(kwargs["path"], "code", classy_repo_rename, "libclass.a")))


def install(path=None, force=False, code=True, no_progress_bars=False, **kwargs):
    log = logging.getLogger(__name__.split(".")[-1])
    if not code:
        log.info("Code not requested. Nothing to do.")
        return True
    log.info("Installing pre-requisites...")
    exit_status = pip_install("cython")
    if exit_status:
        log.error("Could not install pre-requisite: cython")
        return False
    log.info("Downloading classy...")
    success = download_github_release(
        os.path.join(path, "code"), classy_repo_name, classy_repo_version,
        repo_rename=classy_repo_rename, no_progress_bars=no_progress_bars)
    if not success:
        log.error("Could not download classy.")
        return False
    classy_path = os.path.join(path, "code", classy_repo_rename)
    log.info("Compiling classy...")
    from subprocess import Popen, PIPE
    env = deepcopy(os.environ)
    env.update({"PYTHON": sys.executable})
    process_make = Popen(["make"], cwd=classy_path, stdout=PIPE, stderr=PIPE, env=env)
    out, err = process_make.communicate()
    if process_make.returncode:
        log.info(out)
        log.info(err)
        log.error("Compilation failed!")
        return False
    return True



"""
.. module:: _bao_prototype

:Synopsis: BAO, f_sigma8 and other measurements at single redshifts, with correlations
:Author: Antony Lewis (adapted to Cobaya by Jesus Torrado, with little modification)

This code provides a template for BAO, :math:`f\sigma_8`, :math:`H`
and other redshift dependent functions.

The datasets implemented at this moment are:

- ``sixdf_2011_bao``
- ``sdss_dr7_mgs``
- ``sdss_dr12_consensus_bao``
- ``sdss_dr12_consensus_full_shape``
- ``sdss_dr12_consensus_final`` (combined data of the previous two)

.. |br| raw:: html

   <br />

.. note::

   If you use the likelihood ``sixdf_2011_bao``, please cite:
   |br|
   F. Beutler et al,
   `The 6dF Galaxy Survey: baryon acoustic oscillations and the local Hubble constant`
   `(arXiv:1106.3366) <https://arxiv.org/abs/1106.3366>`_
   |br| |br|
   If you use the likelihood ``sdss_dr7_mgs``, please cite:
   |br|
   A.J. Ross et al,
   `The clustering of the SDSS DR7 main Galaxy sample - I.
   A 4 per cent distance measure at z = 0.15`
   `(arXiv:1409.3242) <https://arxiv.org/abs/1409.3242>`_
   |br| |br|
   If you use any of the likelihoods ``sdss_dr12_*``, please cite:
   |br|
   S. Alam et al,
   `The clustering of galaxies in the completed SDSS-III Baryon Oscillation Spectroscopic
   Survey: cosmological analysis of the DR12 galaxy sample`
   `(arXiv:1607.03155) <https://arxiv.org/abs/1607.03155>`_


Usage
-----

To use any of these likelihoods, simply mention them in the likelihoods block, or add them
using the :doc:`input generator <cosmo_basic_runs>`.

These likelihoods have no nuisance parameters or particular settings that you may want
to change (except for the installation path; see below)


Defining your own BAO likelihood
--------------------------------

You can use the likelihood ``bao_generic`` as a template for any BAO data.

To do that, create a file containing the data points, e.g. ``myBAO.dat``, as

.. code::

   [z] [value at z] [quantity]
   ...

where you can use as many different quantities and redshifts as you like.

The available quantities are

- ``DV_over_rs``: Spherically-averaged distance, over sound horizon radius
- ``rs_over_DV``: Idem, inverse
- ``DM_over_rs``: Comoving angular diameter distance, over sound horizon radius
- ``DA_over_rs``: Physical angular diameter distance, over sound horizon radius
- ``Hz_rs``: Hubble parameter, times sound horizon radius
- ``f_sigma8``: Differential matter linear growth rate,
  times amplitude of present-day fuctuations
- ``F_AP``: Anisotropy (Alcock-Paczynski) parameter

In addition create a file, e.g. ``myBAO.cov``, containing the covariance matrix for those
data, with the same row order as the data file.

Now, add to your likelihood block:

.. literalinclude:: ../cobaya/likelihoods/bao_generic/bao_generic.yaml
   :language: yaml


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

Manual installation of the BAO likelihood data
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Assuming you are installing all your
likelihoods under ``/path/to/likelihoods``, simply do

.. code:: bash

   $ cd /path/to/likelihoods
   $ git clone https://github.com/CobayaSampler/bao_data.git

After this, mention the path to this likelihood when you include it in an input file, e.g.

.. code-block:: yaml

   likelihood:
     sdss_dr12_consensus_[bao|full_shape|final|...]:
       path: /path/to/likelihoods/bao_data


"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Global
import os
import numpy as np
from scipy.interpolate import UnivariateSpline
import pandas as pd
import logging

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import HandledException
from cobaya.conventions import _path_install, _c_km_s


class _bao_prototype(Likelihood):

    def initialize(self):
        if not getattr(self, "path", None) and not getattr(self, "path_install", None):
            self.log.error("No path given to BAO data. Set the likelihood property "
                           "'path' or the common property '%s'.", _path_install)
            raise HandledException
        # If no path specified, use the modules path
        data_file_path = os.path.normpath(getattr(self, "path", None) or
                                          os.path.join(self.path_install, "data"))
        # Rescaling by a fiducial value of the sound horizon
        if not hasattr(self, "rs_rescale"):
            if hasattr(self, "rs_fid"):
                self.rs_rescale = 1 / self.rs_fid
            else:
                self.rs_rescale = 1
        # Load "measurements file" and covmat of requested
        if hasattr(self, "measurements_file"):
            try:
                self.data = pd.read_csv(
                    os.path.join(data_file_path, self.measurements_file),
                    header=None, index_col=None, sep="\s+", comment="#")
            except IOError:
                self.log.error("Couldn't find measurements file '%s' in folder '%s'. " % (
                    self.measurements_file, data_file_path) + "Check your paths.")
                raise HandledException
        else:
            self.data = pd.DataFrame([self.data] if not hasattr(self.data[0], "__len__")
                                     else self.data)
        # Colums: z value [err] [type]
        self.has_type = self.data.iloc[:, -1].dtype == np.dtype("O")
        assert self.has_type  # mandatory for now!
        self.has_err = len(self.data.columns) > 2 and self.data[2].dtype == np.float
        if self.has_err:
            self.data.columns = ["z", "value", "error", "observable"]
        else:
            self.data.columns = ["z", "value", "observable"]
        prefix = "bao_"
        self.data["observable"] = [(c[len(prefix):] if c.startswith(prefix) else c)
                                   for c in self.data["observable"]]
        # Probability distribution
        if hasattr(self, "prob_dist"):
            try:
                chi2 = np.loadtxt(os.path.join(data_file_path, self.prob_dist))
            except IOError:
                self.log.error("Couldn't find probability distribution file '%s' "
                               "in folder '%s'. " % (self.prob_dist, data_file_path) +
                               "Check your paths.")
                raise HandledException
            try:
                alpha = np.linspace(
                    self.prob_dist_bounds[0], self.prob_dist_bounds[1], len(chi2))
            except (TypeError, AttributeError, IndexError, ValueError):
                self.log.error("If 'prob_dist' given, 'prob_dist_bounds' needs to be "
                               "specified as [min, max].")
                raise HandledException
            spline = UnivariateSpline(alpha, -chi2 / 2, s=0)
            self.logpdf = lambda x: (
                spline(x)[0] if self.prob_dist_bounds[0] <= x <= self.prob_dist_bounds[1]
                else -np.inf)
        # Covariance --> read and re-sort as self.data
        else:
            try:
                if hasattr(self, "cov_file"):
                    self.cov = np.loadtxt(os.path.join(data_file_path, self.cov_file))
                elif hasattr(self, "invcov_file"):
                    invcov = np.loadtxt(os.path.join(data_file_path, self.invcov_file))
                    self.cov = np.linalg.inv(invcov)
                elif "error" in self.data.columns:
                    self.cov = np.diag(self.data["error"] ** 2)
                else:
                    self.log.error("No errors provided, either as cov, invcov "
                                   "or as the 3rd column in the data file.")
                    raise HandledException
                self.invcov = np.linalg.inv(self.cov)
            except IOError:
                self.log.error("Couldn't find (inv)cov file '%s' in folder '%s'. " % (
                    getattr(self, "cov_file", getattr(self, "invcov_file", None)),
                    data_file_path) + "Check your paths.")
                raise HandledException
            self.logpdf = lambda x: (lambda x_: -0.5 * x_.dot(self.invcov).dot(x_))(
                x - self.data["value"].values)

    def add_theory(self):
        if self.theory.__class__ == "classy":
            self.log.error(
                "BAO likelihood not yet compatible with CLASS (help appreciated!)")
            raise HandledException
        # Requisites
        zs = {obs: self.data.loc[self.data["observable"] == obs, "z"].values
              for obs in self.data["observable"].unique()}
        theory_reqs = {
            "DV_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DV_over_rs", None)},
                "H": {"z": zs.get("DV_over_rs", None)},
                "rdrag": None},
            "rs_over_DV": {
                "angular_diameter_distance": {"z": zs.get("rs_over_DV", None)},
                "H": {"z": zs.get("rs_over_DV", None)},
                "rdrag": None},
            "DM_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DM_over_rs", None)},
                "rdrag": None},
            "DA_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DA_over_rs", None)},
                "rdrag": None},
            "Hz_rs": {
                "H": {"z": zs.get("Hz_rs", None)},
                "rdrag": None},
            "f_sigma8": {
                "fsigma8": {"z": zs.get("f_sigma8", None)},
                "H": {"z": zs.get("Hz_rs", None)}},
            "F_AP": {
                "angular_diameter_distance": {"z": zs.get("F_AP", None)},
                "H": {"z": zs.get("F_AP", None)}}}
        obs_used_not_implemented = np.unique([obs for obs in self.data["observable"]
                                              if obs not in theory_reqs])
        if len(obs_used_not_implemented):
            self.log.error("This likelihood refers to observables '%s' that have not been"
                           " implemented yet. Did you mean any of %s? "
                           "If you didn't, please, open an issue in github.",
                           obs_used_not_implemented, list(theory_reqs))
            raise HandledException
        requisites = {}
        if self.has_type:
            for obs in self.data["observable"].unique():
                requisites.update(theory_reqs[obs])
        self.theory.needs(**requisites)

    def theory_fun(self, z, observable):
        # Functions to get the corresponding theoretical prediction:
        # Spherically-averaged distance, over sound horizon radius
        if observable == "DV_over_rs":
            return np.cbrt(
                ((1 + z) * self.theory.get_angular_diameter_distance(z)) ** 2 *
                _c_km_s * z / self.theory.get_H(z, units="km/s/Mpc")) / self.rs()
        # Idem, inverse
        elif observable == "rs_over_DV":
            return np.cbrt(
                ((1 + z) * self.theory.get_angular_diameter_distance(z)) ** 2 *
                _c_km_s * z / self.theory.get_H(z, units="km/s/Mpc")) ** (-1) * self.rs()
        # Comoving angular diameter distance, over sound horizon radius
        elif observable == "DM_over_rs":
            return (1 + z) * self.theory.get_angular_diameter_distance(z) / self.rs()
        # Physical angular diameter distance, over sound horizon radius
        elif observable == "DA_over_rs":
            return self.theory.get_angular_diameter_distance(z) / self.rs()
        # Hubble parameter, times sound horizon radius
        elif observable == "Hz_rs":
            return self.theory.get_H(z, units="km/s/Mpc") * self.rs()
        # Diff Linear Growth Rate times present amplitude
        elif observable == "f_sigma8":
            return self.theory.get_fsigma8(z)
        # Anisotropy (Alcock-Paczynski) parameter
        elif observable == "F_AP":
            return ((1 + z) * self.theory.get_angular_diameter_distance(z) *
                    self.theory.get_H(z, units="km/s/Mpc")) / _c_km_s

    def rs(self):
        return self.theory.get_param("rdrag") * self.rs_rescale

    def logp(self, **params_values):
        theory = np.array([self.theory_fun(z, obs) for z, obs
                           in zip(self.data["z"], self.data["observable"])]).T[0]
        if self.log.getEffectiveLevel() == logging.DEBUG:
            for i, (z, obs, theo) in enumerate(
                    zip(self.data["z"], self.data["observable"], theory)):
                self.log.debug("%s at z=%g : %g (theo) ; %g (data)",
                               obs, z, theo, self.data.iloc[i, 1])
        return self.logpdf(theory)



"""
.. module:: _cosmo

:Synopsis: Template for Cosmological theory codes.
           Mostly here to document how to compute and get observables.
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
import numpy as np
from copy import deepcopy
from scipy.interpolate import RectBivariateSpline
from six import string_types
from itertools import chain

# Local
from cobaya.theory import Theory
from cobaya.log import HandledException


class _cosmo(Theory):

    def needs(self, **requirements):
        """
        Specifies the quantities that each likelihood needs from the Cosmology code.

        Typical requisites in Cosmology (as keywords, case insensitive):

        - ``Cl={[...]}``: CMB lensed power spectra, as a dictionary ``{spectrum:l_max}``,
          where the possible spectra are combinations of "t", "e", "b" and "p"
          (lensing potential). Get with :func:`~_cosmo.get_cl`.
        - ``H={'z': [z_1, ...], 'units': '1/Mpc' or 'km/s/Mpc'}``: Hubble
          rate at the redshifts requested, in the given units. Get it with
          :func:`~_cosmo.get_H`.
        - ``angular_diameter_distance={'z': [z_1, ...]}``: Physical angular
          diameter distance to the redshifts requested. Get it with
          :func:`~_cosmo.get_angular_diameter_distance`.
        - ``comoving_radial_distance={'z': [z_1, ...]}``: Comoving radial distance
          from us to the redshifts requested. Get it with
          :func:`~_cosmo.get_comoving_radial_distance`.
        - ``fsigma8={'z': [z_1, ...]}``: Structure growth rate
          :math:`f\sigma_8` at the redshifts requested. Get it with
          :func:`~_cosmo.get_fsigma8`.
        - ``k_max=[...]``: Fixes the maximum comoving wavenumber considered.
        - **Other derived parameters** that are not included in the input but whose
          value the likelihood may need.
        """
        if not getattr(self, "_needs", None):
            self._needs = dict([[p, None] for p in self.output_params])
        # TO BE DEPRECATED IN >=1.3
        for product, capitalization in {
                "cl": "Cl", "pk_interpolator": "Pk_interpolator"}.items():
            if product in requirements:
                self.log.error("You requested product '%s', which from now on should be "
                               "capitalized as '%s'.", product, capitalization)
                raise HandledException
        # Accumulate the requirements across several calls in a safe way;
        # e.g. take maximum of all values of a requested precision paramater
        for k, v in requirements.items():
            # Products and other computations
            if k == "Cl":
                self._needs["Cl"] = {
                    cl: max(self._needs.get("Cl", {}).get(cl, 0), v.get(cl, 0))
                    for cl in set(self._needs.get("Cl", {})).union(v)}
            elif k == "Pk_interpolator":
                # Make sure vars_pairs is a list of [list of 2 vars pairs]
                vars_pairs = v.pop("vars_pairs", [])
                try:
                    if isinstance(vars_pairs[0], string_types):
                        vars_pairs = [vars_pairs]
                except IndexError:
                    # Empty list: by default [delta_tot, delta_tot]
                    vars_pairs = [2*["delta_tot"]]
                except:
                    self.log("Cannot understands vars_pairs '%r' for P(k) interpolator",
                             vars_pairs)
                    raise HandledException
                vars_pairs = set([tuple(pair) for pair in chain(
                    self._needs.get(k, {}).get("vars_pairs", []), vars_pairs)])
                self._needs[k] = {
                    "z": np.unique(np.concatenate(
                        (self._needs.get(k, {}).get("z", []),
                         np.atleast_1d(v["z"])))),
                    "k_max": max(
                        self._needs.get(k, {}).get("k_max", 0), v["k_max"]),
                    "vars_pairs": vars_pairs}
                self._needs[k].update(v)
            elif k in ["H", "angular_diameter_distance",
                               "comoving_radial_distance", "fsigma8"]:
                if not k in self._needs:
                    self._needs[k] = {}
                self._needs[k]["z"] = np.unique(np.concatenate(
                    (self._needs[k].get("z", []), v["z"])))
            # Extra derived paramaters and other unknown stuff (keep capitalization)
            elif v is None:
                self._needs[k] = None
            else:
                self.log.error("Unknown required product: '%s:%s'.", k, v)
                raise HandledException

    def requested(self):
        """
        Returns the full set of cosmological products and parameters requested by the
        likelihoods.
        """
        return self._needs

    def get_param(self, p):
        """
        Interface function for likelihoods to get sampled and derived parameters.

        Always use this one; don't try to access theory code attributes directly!
        """
        pass

    def get_cl(self, ell_factor=False, units="muK2"):
        """
        Returns a dictionary of lensed CMB power spectra and the lensing potential ``pp`` power spectrum.

        Set the units with the keyword ``units='1'|'muK2'|'K2'`` (default: 'muK2',
        except for the lensing potential power spectrum, which is always unitless).

        If ``ell_factor=True`` (default: False), multiplies the spectra by
        :math:`\ell(\ell+1)/(2\pi)` (or by :math:`\ell^2(\ell+1)^2/(2\pi)` in the case of
        the lensing potential ``pp`` spectrum).
        """
        pass

    def get_H(self, z, units="km/s/Mpc"):
        """
        Returns the Hubble rate at the given redshifts.

        The redshifts must be a subset of those requested when :func:`~_cosmo.needs`
        was called.

        The available units are ``km/s/Mpc`` (i.e. ``c*H(Mpc^-1)``) and ``1/Mpc``.
        """
        pass

    def get_angular_diameter_distance(self, z):
        """
        Returns the physical angular diameter distance to the given redshifts.

        The redshifts must be a subset of those requested when :func:`~_cosmo.needs`
        was called.
        """
        pass

    def get_Pk_interpolator(self, z):
        """
        Returns a (dict of) power spectrum interpolator(s)
        :class:`PowerSpectrumInterpolator`.
        """
        pass

    def get_fsigma8(self, z):
        """
        Structure growth rate :math:`f\sigma_8`, as defined in eq. 33 of
        `Planck 2015 results. XIII. Cosmological parameters <https://arxiv.org/pdf/1502.01589.pdf>`_,
        at the given redshifts.

        The redshifts must be a subset of those requested when :func:`~_cosmo.needs`
        was called.
        """
        pass

    def get_fsigma8(self, z):
        """
        Structure growth rate :math:`f\sigma_8`, as defined in eq. 33 of
        `Planck 2015 results. XIII. Cosmological parameters <https://arxiv.org/pdf/1502.01589.pdf>`_,
        at the given redshifts.

        The redshifts must be a subset of those requested when :func:`~_cosmo.needs`
        was called.
        """
        pass

    def get_auto_covmat(self, params_info, likes_info):
        """
        Tries to get match to a database of existing covariance matrix files for the current model and data.

        ``params_info`` should contain preferably the slow parameters only.
        """
        from cobaya.cosmo_input import get_best_covmat
        return get_best_covmat(self.path_install, params_info, likes_info)


class PowerSpectrumInterpolator(RectBivariateSpline):
    r"""
    2D spline interpolation object (scipy.interpolate.RectBivariateSpline)
    to evaluate matter power spectrum as function of z and k.

    *This class is adapted from CAMB's own P(k) interpolator, by Antony Lewis;
    it's mostly interface-compatible with the original.*

    :param z: values of z for which the power spectrum was evaluated.
    :param k: values of k for which the power spectrum was evaluated.
    :param P_or_logP: Values of the power spectrum (or log-values, if logP=True).
    :param logk: if True (default: False), assumes that k's are log-spaced.
    :param logP: if True (default: False), log of power spectrum are given and used
        for the underlying interpolator.
    :param extrap_kmax: if set, use power law extrapolation beyond kmax up to
        extrap_kmax; useful for tails of integrals.
    """

    def __init__(self, z, k, P_or_logP, extrap_kmax=None, logk=False, logP=False):
        self.logk, self.logP = logk, logP
        #  Check order
        z, k = [np.atleast_1d(x) for x in [z,k]]
        i_z = np.argsort(z)
        i_k = np.argsort(k)
        self.z, self.k, P_or_logP = z[i_z], k[i_k], P_or_logP[i_z, :][:, i_k]
        self.zmin, self.zmax = np.min(self.z), np.max(self.z)
        self._fk = (lambda k: k if logk else np.log(k))
        self._finvk = (lambda k: np.exp(k) if logk else k)
        self.kmin, self.kmax = np.min(self.k), np.max(self.k)
        # Continue until extrap_kmax using a (log,log)-linear extrapolation
        if extrap_kmax and extrap_kmax > self.kmax:
            logknew = np.log(np.hstack([self.k, extrap_kmax]))
            logPnew = np.empty((P_or_logP.shape[0], P_or_logP.shape[1] + 1))
            logPnew[:, :-1] = P_or_logP if self.logP else np.log(P_or_logP)
            logPnew[:, -1] = (
                logPnew[:, -2] +
                (logPnew[:, -2] - logPnew[:, -3]) / (logknew[-2] - logknew[-3]) *
                (logknew[-1] - logknew[-2]))
            self.k, P_or_logP = np.exp(logknew), logPnew if self.logP else np.exp(logPnew)
        super(self.__class__, self).__init__(self.z, self._fk(self.k), P_or_logP)

    def P(self, z, k, grid=None):
        """
        Get the power spectrum at (z,k).
        """
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.logP:
            return np.exp(self.logP(z, k, grid=grid))
        else:
            return self(z, self._fk(k), grid=grid)

    def logP(self, z, k, grid=None):
        """
        Get the log power spectrum at (z,k).
        """
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.logP:
            return self(z, self._fk(k), grid=grid)
        else:
            return self.P(z, k, grid=grid)



"""
.. module:: samplers.polychord

:Synopsis: Interface for the PolyChord nested sampler
:Author: Will Handley, Mike Hobson and Anthony Lasenby (for PolyChord),
         Jesus Torrado (for the cobaya wrapper only)
"""
# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function

# Global
import os
import sys
import numpy as np
import logging
import inspect
from itertools import chain
from collections import OrderedDict as odict

# Local
from cobaya.tools import read_dnumber, get_external_function, relative_to_int
from cobaya.sampler import Sampler
from cobaya.mpi import get_mpi_comm
from cobaya.mpi import am_single_or_primary_process, more_than_one_process, sync_processes
from cobaya.collection import Collection
from cobaya.log import HandledException
from cobaya.install import download_github_release
from cobaya.yaml import yaml_dump_file

clusters = "clusters"


class polychord(Sampler):
    def initialize(self):
        """Imports the PolyChord sampler and prepares its arguments."""
        if am_single_or_primary_process():  # rank = 0 (MPI master) or None (no MPI)
            self.log.info("Initializing")
        # If path not given, try using general path to modules
        if not self.path and self.path_install:
            self.path = get_path(self.path_install)
        if self.path:
            if am_single_or_primary_process():
                self.log.info("Importing *local* PolyChord from " + self.path)
                if not os.path.exists(os.path.realpath(self.path)):
                    self.log.error("The given path does not exist.")
                    raise HandledException
            pc_build_path = get_build_path(self.path)
            if not pc_build_path:
                self.log.error("Either PolyChord is not in the given folder, "
                               "'%s', or you have not compiled it.", self.path)
                raise HandledException
            # Inserting the previously found path into the list of import folders
            sys.path.insert(0, pc_build_path)
        else:
            self.log.info("Importing *global* PolyChord.")
        try:
            import pypolychord
            from pypolychord.settings import PolyChordSettings
            self.pc = pypolychord
        except ImportError:
            self.log.error(
                "Couldn't find the PolyChord python interface. "
                "Make sure that you have compiled it, and that you either\n"
                " (a) specify a path (you didn't) or\n"
                " (b) install the Python interface globally with\n"
                "     '/path/to/PolyChord/python setup.py install --user'")
            raise HandledException
        # Prepare arguments and settings
        self.nDims = self.model.prior.d()
        self.nDerived = (len(self.model.parameterization.derived_params()) +
                         len(self.model.prior) + len(self.model.likelihood._likelihoods))
        if self.logzero is None:
            self.logzero = np.nan_to_num(-np.inf)
        if self.max_ndead == np.inf:
            self.max_ndead = -1
        for p in ["nlive", "num_repeats", "nprior", "max_ndead"]:
            setattr(self, p, read_dnumber(getattr(self, p), self.nDims, dtype=int))
        # Fill the automatic ones
        if getattr(self, "feedback", None) is None:
            values = {logging.CRITICAL: 0, logging.ERROR: 0, logging.WARNING: 0,
                      logging.INFO: 1, logging.DEBUG: 2}
            self.feedback = values[self.log.getEffectiveLevel()]
        try:
            output_folder = getattr(self.output, "folder")
            output_prefix = getattr(self.output, "prefix") or ""
            self.read_resume = self.resuming
        except AttributeError:
            # dummy output -- no resume!
            self.read_resume = False
            from tempfile import gettempdir
            output_folder = gettempdir()
            if am_single_or_primary_process():
                from random import random
                output_prefix = hex(int(random() * 16 ** 6))[2:]
            else:
                output_prefix = None
            if more_than_one_process():
                output_prefix = get_mpi_comm().bcast(output_prefix, root=0)
        self.base_dir = os.path.join(output_folder, self.base_dir)
        self.file_root = output_prefix
        if am_single_or_primary_process():
            # Creating output folder, if it does not exist (just one process)
            if not os.path.exists(self.base_dir):
                os.makedirs(self.base_dir)
            # Idem, a clusters folder if needed -- notice that PolyChord's default
            # is "True", here "None", hence the funny condition below
            if self.do_clustering is not False:  # None here means "default"
                try:
                    os.makedirs(os.path.join(self.base_dir, clusters))
                except OSError:  # exists!
                    pass
            self.log.info("Storing raw PolyChord output in '%s'.",
                          self.base_dir)
        # Exploiting the speed hierarchy
        if self.blocking:
            speeds, blocks = self.model.likelihood._check_speeds_of_params(self.blocking)
            # Speeds need to be integer to be interpreted as # steps per block
            speeds = relative_to_int(speeds)
            # Account for num_repeats, which is ignored when blocking set by hand:
            # -- multiply the number of steps in the slowest block by num_repeats
            speeds *= self.num_repeats
        else:
            speeds, blocks = self.model.likelihood._speeds_of_params(int_speeds=True)
        blocks_flat = list(chain(*blocks))
        self.ordering = [
            blocks_flat.index(p) for p in self.model.parameterization.sampled_params()]
        self.grade_dims = np.array([len(block) for block in blocks])
        self.grade_frac = np.array(speeds)*self.grade_dims  # steps per block
        # bugfix: pypolychord's C interface for Fortran does not like int numpy types
        self.grade_dims = [int(x) for x in self.grade_dims]
        # Assign settings
        pc_args = ["nlive", "num_repeats", "nprior", "do_clustering",
                   "precision_criterion", "max_ndead", "boost_posterior", "feedback",
                   "logzero", "posteriors", "equals", "compression_factor",
                   "cluster_posteriors", "write_resume", "read_resume", "write_stats",
                   "write_live", "write_dead", "base_dir", "grade_frac", "grade_dims",
                   "feedback", "read_resume", "base_dir", "file_root", "grade_frac",
                   "grade_dims"]
        self.pc_settings = PolyChordSettings(
            self.nDims, self.nDerived, seed=(self.seed if self.seed is not None else -1),
            **{p: getattr(self, p) for p in pc_args if getattr(self, p) is not None})
        # prior conversion from the hypercube
        bounds = self.model.prior.bounds(
            confidence_for_unbounded=self.confidence_for_unbounded)
        # Check if priors are bounded (nan's to inf)
        inf = np.where(np.isinf(bounds))
        if len(inf[0]):
            params_names = self.model.parameterization.sampled_params()
            params = [params_names[i] for i in sorted(list(set(inf[0])))]
            self.log.error("PolyChord needs bounded priors, but the parameter(s) '"
                           "', '".join(params) + "' is(are) unbounded.")
            raise HandledException
        locs = bounds[:, 0]
        scales = bounds[:, 1] - bounds[:, 0]
        # This function re-scales the parameters AND puts them in the right order
        self.pc_prior = lambda x: (locs + np.array(x)[self.ordering] * scales).tolist()
        # We will need the volume of the prior domain, since PolyChord divides by it
        self.logvolume = np.log(np.prod(scales))
        # Prepare callback function
        if self.callback_function is not None:
            self.callback_function_callable = (
                get_external_function(self.callback_function))
        self.last_point_callback = 0
        # Prepare runtime live and dead points collections
        self.live = Collection(
            self.model, None, name="live", initial_size=self.pc_settings.nlive)
        self.dead = Collection(self.model, self.output, name="dead")
        self.n_sampled = len(self.model.parameterization.sampled_params())
        self.n_derived = len(self.model.parameterization.derived_params())
        self.n_priors = len(self.model.prior)
        self.n_likes = len(self.model.likelihood._likelihoods)
        # Done!
        if am_single_or_primary_process():
            self.log.info("Calling PolyChord with arguments:")
            for p, v in inspect.getmembers(self.pc_settings, lambda a: not (callable(a))):
                if not p.startswith("_"):
                    self.log.info("  %s: %s", p, v)

    def dumper(self, live_points, dead_points, logweights, logZ, logZstd):
        # Store live and dead points and evidence computed so far
        self.live.reset()
        for point in live_points:
            self.live.add(
                point[:self.n_sampled],
                derived=point[self.n_sampled:self.n_sampled + self.n_derived],
                weight=np.nan,
                logpriors=point[self.n_sampled + self.n_derived:
                                self.n_sampled + self.n_derived + self.n_priors],
                loglikes=point[self.n_sampled + self.n_derived + self.n_priors:
                               self.n_sampled + self.n_derived + self.n_priors + self.n_likes])
        for logweight, point in zip(logweights[self.last_point_callback:],
                                    dead_points[self.last_point_callback:]):
            self.dead.add(
                point[:self.n_sampled],
                derived=point[self.n_sampled:self.n_sampled + self.n_derived],
                weight=np.exp(logweight),
                logpriors=point[self.n_sampled + self.n_derived:
                                self.n_sampled + self.n_derived + self.n_priors],
                loglikes=point[self.n_sampled + self.n_derived + self.n_priors:
                               self.n_sampled + self.n_derived + self.n_priors + self.n_likes])
        self.logZ, self.logZstd = logZ, logZstd
        # Callback function
        if self.callback_function is not None:
            self.callback_function_callable(self)
            self.last_point_callback = self.dead.n()

    def run(self):
        """
        Prepares the posterior function and calls ``PolyChord``'s ``run`` function.
        """

        # Prepare the posterior
        # Don't forget to multiply by the volume of the physical hypercube,
        # since PolyChord divides by it
        def logpost(params_values):
            logposterior, logpriors, loglikes, derived = (
                self.model.logposterior(params_values))
            if len(derived) != len(self.model.parameterization.derived_params()):
                derived = np.full(
                    len(self.model.parameterization.derived_params()), np.nan)
            if len(loglikes) != len(self.model.likelihood._likelihoods):
                loglikes = np.full(
                    len(self.model.likelihood._likelihoods), np.nan)
            derived = list(derived) + list(logpriors) + list(loglikes)
            return (
                max(logposterior + self.logvolume, 0.99 * self.pc_settings.logzero), derived)

        sync_processes()
        if am_single_or_primary_process():
            self.log.info("Sampling!")
        self.pc.run_polychord(logpost, self.nDims, self.nDerived, self.pc_settings,
                              self.pc_prior, self.dumper)

    def save_sample(self, fname, name):
        sample = np.atleast_2d(np.loadtxt(fname))
        if not sample.size:
            return None
        collection = Collection(self.model, self.output, name=str(name))
        for row in sample:
            collection.add(
                row[2:2 + self.n_sampled],
                derived=row[2 + self.n_sampled:2 + self.n_sampled + self.n_derived],
                weight=row[0],
                logpriors=row[-(self.n_priors + self.n_likes):-self.n_likes],
                loglikes=row[-self.n_likes:])
        # make sure that the points are written
        collection._out_update()
        return collection

    def close(self, exception_type=None, exception_value=None, traceback=None):
        """
        Loads the sample of live points from ``PolyChord``'s raw output and writes it
        (if ``txt`` output requested).
        """
        if exception_type:
            raise
        if am_single_or_primary_process():
            self.log.info("Loading PolyChord's results: samples and evidences.")
            prefix = os.path.join(self.pc_settings.base_dir, self.pc_settings.file_root)
            self.collection = self.save_sample(prefix + ".txt", "1")
            if self.pc_settings.do_clustering is not False:  # NB: "None" == "default"
                self.clusters = {}
                do_output = hasattr(self.output, "folder")
                for f in os.listdir(os.path.join(self.pc_settings.base_dir, clusters)):
                    if not f.startswith(self.pc_settings.file_root):
                        continue
                    if do_output:
                        cluster_folder = os.path.join(
                            self.output.folder, self.output.prefix +
                                                ("_" if self.output.prefix else "") + clusters)
                        if not os.path.exists(cluster_folder):
                            os.mkdir(cluster_folder)
                    try:
                        i = int(f[len(self.pc_settings.file_root) + 1:-len(".txt")])
                    except ValueError:
                        continue
                    if do_output:
                        old_folder = self.output.folder
                        self.output.folder = cluster_folder
                    fname = os.path.join(self.pc_settings.base_dir, clusters, f)
                    sample = self.save_sample(fname, str(i))
                    self.clusters[i] = {"sample": sample}
                    if do_output:
                        self.output.folder = old_folder
            # Prepare the evidence(s) and write to file
            pre = "log(Z"
            active = "(Still active)"
            lines = []
            with open(prefix + ".stats", "r") as statsfile:
                lines = [l for l in statsfile.readlines() if l.startswith(pre)]
            for l in lines:
                logZ, logZstd = [float(n.replace(active, "")) for n in
                                 l.split("=")[-1].split("+/-")]
                component = l.split("=")[0].lstrip(pre + "_").rstrip(") ")
                if not component:
                    self.logZ, self.logZstd = logZ, logZstd
                elif self.pc_settings.do_clustering:
                    i = int(component)
                    self.clusters[i]["logZ"], self.clusters[i]["logZstd"] = logZ, logZstd
            if do_output:
                out_evidences = odict([["logZ", self.logZ], ["logZstd", self.logZstd]])
                if self.clusters:
                    out_evidences["clusters"] = odict()
                    for i in sorted(list(self.clusters.keys())):
                        out_evidences["clusters"][i] = odict(
                            [["logZ", self.clusters[i]["logZ"]],
                             ["logZstd", self.clusters[i]["logZstd"]]])
                fname = os.path.join(self.output.folder, self.output.prefix+".logZ")
                yaml_dump_file(fname, out_evidences, comment="log-evidence", error_if_exists=False)
        # TODO: try to broadcast the collections
        #        if get_mpi():
        #            bcast_from_0 = lambda attrname: setattr(self,
        #                attrname, get_mpi_comm().bcast(getattr(self, attrname, None), root=0))
        #            map(bcast_from_0, ["collection", "logZ", "logZstd", "clusters"])
        if am_single_or_primary_process():
            self.log.info("Finished! Raw PolyChord output stored in '%s', "
                          "with prefix '%s'",
                          self.pc_settings.base_dir, self.pc_settings.file_root)

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``Collection`` containing the sequentially discarded live points.
        """
        if am_single_or_primary_process():
            products = {
                "sample": self.collection, "logZ": self.logZ, "logZstd": self.logZstd}
            if self.pc_settings.do_clustering:
                products.update({"clusters": self.clusters})
            return products
        else:
            return {}


# Installation routines ##################################################################

# Name of the PolyChord repo and version to download
pc_repo_name = "PolyChord/PolyChordLite"
pc_repo_version = "1.16"


def get_path(path):
    return os.path.realpath(
        os.path.join(path, "code", pc_repo_name[pc_repo_name.find("/")+1:]))


def get_build_path(polychord_path):
    try:
        build_path = os.path.join(polychord_path, "build")
        post = next(d for d in os.listdir(build_path) if d.startswith("lib."))
        build_path = os.path.join(build_path, post)
        return build_path
    except (OSError, StopIteration):
        return False


def is_installed(**kwargs):
    if not kwargs["code"]:
        return True
    poly_path = get_path(kwargs["path"])
    if not os.path.isfile(os.path.realpath(os.path.join(poly_path, "lib/libchord.so"))):
        return False
    poly_build_path = get_build_path(poly_path)
    if not poly_build_path:
        return False
    sys.path.insert(0, poly_build_path)
    try:
        import pypolychord
        return True
    except ImportError:
        return False


def install(path=None, force=False, code=False, data=False, no_progress_bars=False):
    if not code:
        return True
    log = logging.getLogger(__name__.split(".")[-1])
    log.info("Downloading PolyChord...")
    success = download_github_release(os.path.join(path, "code"), pc_repo_name,
                                      pc_repo_version, no_progress_bars=no_progress_bars)
    if not success:
        log.error("Could not download PolyChord.")
        return False
    log.info("Compiling (Py)PolyChord...")
    from subprocess import Popen, PIPE
    # Needs to re-define os' PWD,
    # because MakeFile calls it and is not affected by the cwd of Popen
    cwd = os.path.join(path, "code", pc_repo_name[pc_repo_name.find("/")+1:])
    my_env = os.environ.copy()
    my_env.update({"PWD": cwd})
    process_make = Popen(["make", "pypolychord", "MPI=1"], cwd=cwd, env=my_env,
                         stdout=PIPE, stderr=PIPE)
    out, err = process_make.communicate()
    if process_make.returncode:
        log.info(out)
        log.info(err)
        log.error("Compilation failed!")
        return False
    my_env.update({"CC": "mpicc", "CXX": "mpicxx"})
    process_make = Popen([sys.executable, "setup.py", "build"],
                         cwd=cwd, env=my_env, stdout=PIPE, stderr=PIPE)
    out, err = process_make.communicate()
    if process_make.returncode:
        log.info(out)
        log.info(err)
        log.error("Python build failed!")
        return False
    return True



"""
.. module:: bicep_keck_2015

:Synopsis: Likelihood of BICEP2/Keck-Array, October 2018 (2015 data)
:Author: BICEP2/Keck team and Antony Lewis

.. warning::

   This is still in **beta**. Coming soon!

   For now, use ``bicep_keck_2014`` instead, which was used in the Planck 2015 analysis.

.. |br| raw:: html

   <br />

.. note::

   **If you use this likelihood, please cite it as:**
   |br|
   Keck Array and BICEP2 Collaborations,
   `BICEP2 / Keck Array x: Constraints on Primordial Gravitational Waves
   using Planck, WMAP, and New BICEP2/Keck Observations through the 2015 Season`
   `(arXiv:1810.05216) <https://arxiv.org/abs/1810.05216>`_

Usage
-----

To use this likelihood, ``bicep_keck_2015``, you simply need to mention it in the
``likelihood`` block, or add them using the :doc:`input generator <cosmo_basic_runs>`.

The corresponding nuisance parameters will be added automatically,
so you don't have to care about listing them in the ``params`` block.

The nuisance parameters and their default priors can be found in the ``defaults.yaml``
files in the folder for the source code of this module, and it's reproduced below.

You shouldn't need to modify any of the options of this simple likelihood,
but if you really need to, just copy the ``likelihood`` block into your input ``yaml``
file and modify whatever options you want (you can delete the rest).

.. literalinclude:: ../cobaya/likelihoods/bicep_keck_2015/bicep_keck_2015.yaml
   :language: yaml


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

Manual installation of the BICEP2/Keck-Array likelihood
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Assuming you are installing all your
likelihoods under ``/path/to/likelihoods``, simply do

.. code:: bash

   $ cd /path/to/likelihoods
   $ mkdir bicep_keck_2015
   $ cd bicep_keck_2015
   $ wget http://bicepkeck.org/BK15_datarelease/BK15_cosmomc.tgz
   $ tar xvf BK15_cosmomc.tgz
   $ rm BK15_cosmomc.tgz

After this, mention the path to this likelihood when you include it in an input file as

.. code-block:: yaml

   likelihood:
     bicep_keck_2015:
       path: /path/to/likelihoods/bicep_keck_2015

"""

import os
import numpy as np

# Local
from cobaya.likelihoods._cmblikes_prototype import _cmblikes_prototype
from cobaya.conventions import _h_J_s, _kB_J_K, _T_CMB_K
from cobaya.install import download_file

# Logger
import logging

# Physical constants
Ghz_Kelvin = _h_J_s / _kB_J_K * 1e9


class bicep_keck_2015(_cmblikes_prototype):

    def readIni(self, ini):

        super(self.__class__, self).readIni(ini)
        self.fpivot_dust = ini.float('fpivot_dust', 353.0)
        self.fpivot_sync = ini.float('fpivot_sync', 23.0)
        self.bandpasses = []
        for i, used in enumerate(self.used_map_order):
            self.bandpasses += [
                self.read_bandpass(ini.relativeFileName('bandpass[%s]' % used))]

        self.fpivot_dust_decorr = [ini.array_float('fpivot_dust_decorr', i, default) for i, default in
                                   zip([1, 2], [217., 353.])]
        self.fpivot_sync_decorr = [ini.array_float('fpivot_sync_decorr', i, default) for i, default in
                                   zip([1, 2], [22., 33.])]
        self.lform_dust_decorr = ini.string('lform_dust_decorr', 'flat')
        self.lform_sync_decorr = ini.string('lform_sync_decorr', 'flat')

    def read_bandpass(self, fname):
        bandpass = Bandpass()
        bandpass.R = np.loadtxt(fname)
        nu = bandpass.R[:, 0]
        bandpass.dnu = np.hstack(
            ((nu[1] - nu[0]), (nu[2:] - nu[:-2]) / 2, (nu[-1] - nu[-2])))
        # Calculate thermodynamic temperature conversion between this bandpass
        # and pivot frequencies 353 GHz (usedfor dust) and 150 GHz (used for sync).
        th_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** 4 *
                        np.exp(Ghz_Kelvin * bandpass.R[:, 0] / _T_CMB_K) /
                        (np.exp(Ghz_Kelvin * bandpass.R[:, 0] / _T_CMB_K) - 1) ** 2)
        nu0 = self.fpivot_dust
        th0 = (nu0 ** 4 * np.exp(Ghz_Kelvin * nu0 / _T_CMB_K) /
               (np.exp(Ghz_Kelvin * nu0 / _T_CMB_K) - 1) ** 2)
        bandpass.th_dust = th_int / th0
        nu0 = self.fpivot_sync
        th0 = (nu0 ** 4 * np.exp(Ghz_Kelvin * nu0 / _T_CMB_K) /
               (np.exp(Ghz_Kelvin * nu0 / _T_CMB_K) - 1) ** 2)
        bandpass.th_sync = th_int / th0
        # Calculate bandpass center-of-mass (i.e. mean frequency).
        bandpass.nu_bar = np.dot(bandpass.dnu, bandpass.R[:, 0] * bandpass.R[:, 1]) / np.dot(bandpass.dnu,
                                                                                             bandpass.R[:, 1])

        return bandpass

    def dust_scaling(self, beta, Tdust, bandpass, nu0, bandcenter_err):
        """Calculates greybody scaling of dust signal defined at 353 GHz
        to specified bandpass."""
        gb_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** (3 + beta) /
                        (np.exp(Ghz_Kelvin * bandpass.R[:, 0] / Tdust) - 1))
        # Calculate values at pivot frequency.
        gb0 = nu0 ** (3 + beta) / (np.exp(Ghz_Kelvin * nu0 / Tdust) - 1)
        #  Add correction for band center error
        if bandcenter_err != 1:
            nu_bar = Ghz_Kelvin * bandpass.nu_bar
            # Conversion factor error due to bandcenter error.
            th_err = (bandcenter_err) ** 4 * \
                     np.exp(Ghz_Kelvin * bandpass % nu_bar * (bandcenter_err - 1) / _T_CMB_K) \
                         (np.exp(nu_bar / _T_CMB_K) - 1) ** 2 / \
                     (np.exp(nu_bar * bandcenter_err / _T_CMB_K) - 1) ** 2
            # Greybody scaling error due to bandcenter error.
            gb_err = (bandcenter_err) ** (3 + beta) * \
                     (np.exp(nu_bar / Tdust) - 1) / \
                     (np.exp(nu_bar * bandcenter_err / Tdust) - 1)
        else:
            th_err = 1
            gb_err = 1

        # Calculate dust scaling.
        return (gb_int / gb0) / bandpass.th_dust * (gb_err / th_err)

    def sync_scaling(self, beta, bandpass, nu0, bandcenter_err):
        """Calculates power-law scaling of synchrotron signal defined at 150 GHz
        to specified bandpass."""
        # Integrate power-law scaling and thermodynamic temperature conversion
        # across experimental bandpass.
        pl_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** (2 + beta))
        # Calculate values at pivot frequency.
        pl0 = nu0 ** (2 + beta)
        if bandcenter_err != 1:
            nu_bar = Ghz_Kelvin * bandpass.nu_bar
            th_err = (bandcenter_err) ** 4 * \
                     np.exp(nu_bar * (bandcenter_err - 1) / _T_CMB_K) * \
                     (np.exp(nu_bar / _T_CMB_K) - 1) ** 2 / \
                     (np.exp(nu_bar * bandcenter_err / _T_CMB_K) - 1) ** 2
            pl_err = (bandcenter_err) ** (2 + beta)
        else:
            pl_err = 1
            th_err = 1
        # Calculate sync scaling.
        return (pl_int / pl0) / bandpass.th_sync * (pl_err / th_err)

    def decorrelation(self, delta, nu0, nu1, nupivot, rat, lform):
        # Calculate factor by which foreground (dust or sync) power is decreased
        # for a cross-spectrum between two different frequencies.

        # Decorrelation scales as log^2(nu0/nu1). rat is l/l_pivot
        scl_nu = (np.log(nu0 / nu1) ** 2) / (np.log(nupivot(1) / nupivot(2)) ** 2)
        # Functional form for ell scaling is specified in .dataset file.
        if lform == "flat":
            scl_ell = 1.0
        elif lform == 'lin':
            scl_ell = rat
        elif lform == 'quad':
            scl_ell = rat ** 2
        else:
            scl_ell = 1.0

        # Even for small cval, correlation can become negative for sufficiently large frequency
        # difference or ell value (with linear or quadratic scaling).
        # Following Vansyngel et al, A&A, 603, A62 (2017), we use an exponential function to
        # remap the correlation coefficient on to the range [0,1].
        # We symmetrically extend this function to (non-physical) correlation coefficients
        # greater than 1 -- this is only used for validation tests of the likelihood model.
        if delta > 1:
            return 2.0 - np.exp(np.log(2.0 - delta) * scl_nu * scl_ell)
        else:
            return np.exp(np.log(delta) * scl_nu * scl_ell)

    def add_foregrounds(self, cls, data_params):
        lpivot = 80.0
        Adust = data_params['BBdust']
        Async = data_params['BBsync']
        alphadust = data_params['BBalphadust']
        betadust = data_params['BBbetadust']
        Tdust = data_params['BBTdust']
        alphasync = data_params['BBalphasync']
        betasync = data_params['BBbetasync']
        dustsync_corr = data_params['BBdustsynccorr']
        EEtoBB_dust = data_params['EEtoBB_dust']
        EEtoBB_sync = data_params['EEtoBB_sync']
        delta_dust = data_params['delta_dust']
        delta_sync = data_params['delta_sync']
        gamma_corr = data_params['gamma_corr']  # 13

        # Calculate dust and sync scaling for each map.
        bandcenter_err = np.empty(self.nmaps_required)
        fdust = np.empty(self.nmaps_required)
        fsync = np.empty(self.nmaps_required)
        for i, mapname in enumerate(self.used_map_order):
            # Read and assign values to band center error params
            if '95' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_95'] + 1
            elif '150' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_150'] + 1
            elif '220' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_220'] + 1
            else:
                bandcenter_err[i] = 1
            fdust[i] = self.dust_scaling(
                betadust, Tdust, self.bandpasses[i], self.fpivot_dust, bandcenter_err[i])
            fsync[i] = self.sync_scaling(
                betasync, self.bandpasses[i], self.fpivot_sync, bandcenter_err[i])

        rat = np.arange(self.pcl_lmin, self.pcl_lmax + 1) / lpivot
        dustpow = Adust * rat ** alphadust
        syncpow = Async * rat ** alphasync
        dustsyncpow = (dustsync_corr * np.sqrt(Adust * Async) *
                       rat ** ((alphadust + alphasync) / 2))

        #  Only calculate foreground decorrelation if necessary.
        need_sync_decorr = np.abs(delta_sync - 1) > 1e-5
        need_dust_decorr = np.abs(delta_dust - 1) > 1e-5
        for i in range(self.nmaps_required):
            for j in range(i + 1):
                CL = cls[i, j]
                dust = fdust[i] * fdust[j]
                sync = fsync[i] * fsync[j]
                dustsync = fdust[i] * fsync[j] + fsync[i] * fdust[j]
                EE = CL.theory_ij[0] == 1 and CL.theory_ij[1] == 1
                if EE:
                    # EE spectrum: multiply foregrounds by EE/BB ratio
                    dust *= EEtoBB_dust
                    sync *= EEtoBB_sync
                    dustsync *= np.sqrt(EEtoBB_sync * EEtoBB_dust)

                if need_dust_decorr and i != j:
                    corr_dust = self.decorrelation(delta_dust, self.bandpasses[i].nu_bar * bandcenter_err[i], \
                                                   self.bandpasses[j].nu_bar * bandcenter_err[j],
                                                   self.fpivot_dust_decorr, rat, \
                                                   self.lform_dust_decorr)
                else:
                    corr_dust = 1
                if need_sync_decorr and i != j:
                    corr_sync = self.decorrelation(delta_sync, self.bandpasses[i].nu_bar * bandcenter_err[i], \
                                                   self.bandpasses[j].nu_bar * bandcenter_err[j],
                                                   self.pivot_sync_decorr, rat, \
                                                   self.lform_sync_decorr)
                else:
                    corr_sync = 1
                #  Add foreground model to theory spectrum.
                # NOTE: Decorrelation is not implemented for the dust/sync correlated component.
                #      In BK15, we never turned on correlation and decorrelation parameters
                #       simultaneously.
                CL.CL += dust * dustpow * corr_dust + sync * syncpow * corr_sync + dustsync * dustsyncpow


class Bandpass(object):
    pass


# Installation routines ##################################################################

def get_path(path):
    return os.path.realpath(os.path.join(path, "data", __name__.split(".")[-1]))


def is_installed(**kwargs):
    if kwargs["data"]:
        if not os.path.exists(os.path.join(get_path(kwargs["path"]), "BK15_cosmomc")):
            return False
    return True


def install(path=None, name=None, force=False, code=False, data=True,
            no_progress_bars=False):
    log = logging.getLogger(__name__.split(".")[-1])
    full_path = get_path(path)
    if not os.path.exists(full_path):
        os.makedirs(full_path)
    if not data:
        return True
    log.info("Downloading likelihood data...")
    # Refuses http[S]!  (check again after new release)
    filename = r"http://bicepkeck.org/BK15_datarelease/BK15_cosmomc.tgz"
    if not download_file(filename, full_path, decompress=True, logger=log,
                         no_progress_bars=no_progress_bars):
        return False
    log.info("Likelihood data downloaded and uncompressed correctly.")
    return True



"""
.. module:: output

:Synopsis: Generic output class and output drivers
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
import os
import sys
import six
import traceback
import datetime
from copy import deepcopy

# Local
from cobaya.yaml import yaml_dump, yaml_load, yaml_load_file, OutputError
from cobaya.conventions import _input_suffix, _full_suffix, _separator, _yaml_extensions
from cobaya.conventions import _resume, _resume_default, _force
from cobaya.conventions import _likelihood, _params
from cobaya.log import HandledException
from cobaya.input import is_equal_info
from cobaya.mpi import am_single_or_primary_process, get_mpi_comm

# Logger
import logging


class Output(object):
    def __init__(self, output_prefix=None, resume=_resume_default, force_output=False):
        self.log = logging.getLogger("output")
        self.folder = os.sep.join(output_prefix.split(os.sep)[:-1]) or "."
        self.prefix = (lambda x: x if x != "." else "")(output_prefix.split(os.sep)[-1])
        self.force_output = force_output
        if resume and force_output and output_prefix:
            # No resume and force at the same time (if output)
            self.log.error(
                "Make 'resume: True' or 'force: True', not both at the same time: "
                "can't simultaneously overwrite a chain and resume from it.")
            raise HandledException
        if not os.path.exists(self.folder):
            self.log.debug("Creating output folder '%s'", self.folder)
            try:
                os.makedirs(self.folder)
            except OSError:
                self.log.error("".join(["-"] * 20 + ["\n\n"] +
                                       list(traceback.format_exception(*sys.exc_info())) +
                                       ["\n"] + ["-"] * 37))
                self.log.error("Could not create folder '%s'. "
                               "See traceback on top of this message.", self.folder)
                raise HandledException
        self.log.info("Output to be read-from/written-into folder '%s', with prefix '%s'",
                      self.folder, self.prefix)
        # Prepare file names, and check if chain exists
        info_file_prefix = os.path.join(
            self.folder, self.prefix + (_separator if self.prefix else ""))
        self.file_input = info_file_prefix + _input_suffix + _yaml_extensions[0]
        self.file_full = info_file_prefix + _full_suffix + _yaml_extensions[0]
        self.resuming = False
        if os.path.isfile(self.file_full):
            self.log.info(
                "Found existing products with the requested ouput prefix: '%s'",
                output_prefix)
            if self.force_output:
                self.log.info("Deleting previous chain ('force' was requested).")
                [os.remove(f) for f in [self.file_input, self.file_full]]
            elif resume:
                # Only in this case we can be sure that we are actually resuming
                self.resuming = True
                self.log.info("Let's try to resume/load.")
            else:
                # If only input and full info dumped, overwrite; else fail
                info_files = [
                    os.path.basename(f) for f in [self.file_input, self.file_full]]
                same_prefix_noinfo = [f for f in os.listdir(self.folder) if
                                      f.startswith(self.prefix) and f not in info_files]
                if not same_prefix_noinfo:
                    [os.remove(f) for f in [self.file_input, self.file_full]]
                    self.log.info("Overwritten old failed chain files.")
                else:
                    self.log.error("Delete the previous sample manually, automatically "
                                   "('-%s', '--%s', '%s: True')" % (
                                       _force[0], _force, _force) +
                                   " or request resuming ('-%s', '--%s', '%s: True')" % (
                                       _resume[0], _resume, _resume))
                    raise HandledException
        # Output kind and collection extension
        self.kind = "txt"
        self.ext = "txt"

    def updated_output_prefix(self):
        """
        Updated path: drops folder: now it's relative to the chain's location.
        """
        return self.prefix or "."

    def is_resuming(self):
        return self.resuming

    def reload_full_info(self):
        return yaml_load_file(self.file_full)

    def dump_info(self, input_info, full_info, check_compatible=True):
        """
        Saves the info in the chain folder twice:
           - the input info.
           - idem, populated with the modules' defaults.

        If resuming a sample, checks first that old and new infos are consistent.
        """
        # trim known params of each likelihood: for internal use only
        full_info_trimmed = deepcopy(full_info)
        for lik_info in full_info_trimmed.get(_likelihood, {}).values():
            if hasattr(lik_info, "pop"):
                lik_info.pop(_params, None)
        if check_compatible:
            try:
                # We will test the old info agains the dumped+loaded new info.
                # This is because we can't actually check if python objects do change
                old_info = self.reload_full_info()
                new_info = yaml_load(yaml_dump(full_info_trimmed))
                if not is_equal_info(old_info, new_info, strict=False):
                    self.log.error("Old and new sample information not compatible! "
                                   "Resuming not possible!")
                    raise HandledException
            except IOError:
                # There was no previous chain
                pass
        # We write the new one anyway (maybe updated debug, resuming...)
        for f, info in [(self.file_input, input_info),
                        (self.file_full, full_info_trimmed)]:
            if not info:
                pass
            with open(f, "w") as f_out:
                try:
                    f_out.write(yaml_dump(info))
                except OutputError as e:
                    self.log.error(str(e))
                    raise HandledException

    def prepare_collection(self, name=None, extension=None):
        if not name:
            name = (datetime.datetime.now().isoformat()
                    .replace("T", "_").replace(":", "-").replace(".", "-"))
        file_name = os.path.join(
            self.folder,
            self.prefix + ("_" if self.prefix else "") + name + "." + (extension or self.ext))
        return file_name, self.kind


class OutputDummy(Output):
    """
    Dummy output class. Does nothing. Evaluates to 'False' as a class.
    """

    def __init__(self, *args, **kwargs):
        self.log = logging.getLogger("output")
        self.log.debug("No output requested. Doing nothing.")
        # override all methods that actually produce output
        exclude = ["nullfunc"]
        _func_name = "__name__" if six.PY3 else "func_name"
        for attrname, attr in list(Output.__dict__.items()):
            func_name = getattr(attr, _func_name, None)
            if func_name and func_name not in exclude and '__' not in func_name:
                setattr(self, attrname, self.nullfunc)

    def nullfunc(self, *args, **kwargs):
        pass

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False


class Output_MPI(Output):
    """
    MPI wrapper around the Output class.
    """

    def __init__(self, *args, **kwargs):
        to_broadcast = ("folder", "prefix", "kind", "ext", "resuming")
        if am_single_or_primary_process():
            Output.__init__(self, *args, **kwargs)
        else:
            for var in to_broadcast:
                setattr(self, var, None)
        for var in to_broadcast:
            setattr(self, var, get_mpi_comm().bcast(getattr(self, var), root=0))

    def dump_info(self, *args, **kwargs):
        if am_single_or_primary_process():
            Output.dump_info(self, *args, **kwargs)


def get_Output(*args, **kwargs):
    """
    Auxiliary function to retrieve the output driver.
    """
    if kwargs.get("output_prefix"):
        from cobaya.mpi import import_MPI
        Output = import_MPI(".output", "Output")
        return Output(*args, **kwargs)
    else:
        return OutputDummy(*args, **kwargs)



"""
.. module:: input

:Synopsis: Input-related functions
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function
from functools import reduce

# Global
import os
from collections import OrderedDict as odict
from copy import deepcopy
from importlib import import_module
import inspect
from six import string_types
from itertools import chain
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    # Suppress message about optional dependency
    from fuzzywuzzy import process as fuzzy_process

# Local
from cobaya.conventions import _package, _products_path, _path_install, _resume
from cobaya.conventions import _output_prefix, _debug, _debug_file
from cobaya.conventions import _params, _prior, _theory, _likelihood, _sampler, _external
from cobaya.conventions import _p_label, _p_derived, _p_ref, _p_drop, _p_value, _p_renames
from cobaya.conventions import _p_proposal, _input_params, _output_params
from cobaya.tools import get_folder, recursive_update, recursive_odict_to_dict
from cobaya.yaml import yaml_load_file
from cobaya.log import HandledException
from cobaya.parameterization import expand_info_param
from cobaya.mpi import get_mpi_comm, am_single_or_primary_process

# Logger
import logging

log = logging.getLogger(__name__.split(".")[-1])


def load_input(input_file):
    """
    Loads general info, and splits it into the right parts.
    """
    file_name, extension = os.path.splitext(input_file)
    file_name = os.path.basename(file_name)
    if extension not in (".yaml", ".yml"):
        log.error("Extension of input file '%s' not recognized.", input_file)
        raise HandledException
    info = yaml_load_file(input_file) or {}
    # if output_prefix not defined, default to input_file name (sans ext.) as prefix;
    if _output_prefix not in info:
        info[_output_prefix] = file_name
    # warn if no output, since we are in shell-invocation mode.
    elif info[_output_prefix] is None:
        log.warning("WARNING: Output explicitly suppressed with 'output_prefix: null'")
    # contained? Ensure that output is sent where it should
    if "CONTAINED" in os.environ:
        for out in [_output_prefix, _debug_file]:
            if info.get(out):
                if not info[out].startswith("/"):
                    info[out] = os.path.join(_products_path, info[out])
    return info


# MPI wrapper for loading the input info
def load_input_MPI(input_file):
    if am_single_or_primary_process():
        info = load_input(input_file)
    else:
        info = None
    info = get_mpi_comm().bcast(info, root=0)
    return info


def get_modules(*infos):
    """Returns modules all requested as an odict ``{kind: set([modules])}``.
    Priors are not included."""
    modules = odict()
    for info in infos:
        for field in [_theory, _likelihood, _sampler]:
            if field not in modules:
                modules[field] = []
            modules[field] += [a for a in (info.get(field) or [])
                               if a not in modules[field]]
    # pop empty blocks
    for k, v in list(modules.items()):
        if not v:
            modules.pop(k)
    return modules


def get_default_info(module, kind):
    path_to_defaults = os.path.join(get_folder(module, kind), module + ".yaml")
    try:
        default_module_info = yaml_load_file(path_to_defaults)
    except IOError:
        # probably an external module
        default_module_info = {kind: {module: {}}}
        log.debug("Module %s:%s does not have a defaults file. " % (kind, module) +
                  "Maybe it is an external module.")
    try:
        default_module_info[kind][module]
    except KeyError:
        log.error("The defaults file for '%s' should be structured "
                  "as %s:%s:{[options]}.", module, kind, module)
        raise HandledException
    return default_module_info


def get_full_info(info):
    """
    Creates an updated info starting from the defaults for each module and updating it
    with the input info.
    """
    # Don't modify the original input!
    input_info = deepcopy(info)
    # Creates an equivalent info using only the defaults
    full_info = odict()
    default_params_info = odict()
    default_prior_info = odict()
    modules = get_modules(input_info)
    for block in modules:
        full_info[block] = odict()
        for module in modules[block]:
            # Start with the default class options
            full_info[block][module] = deepcopy(getattr(
                import_module(_package + "." + block, package=_package), "class_options", {}))
            # Go on with defaults
            default_module_info = get_default_info(module, block)
            full_info[block][module].update(default_module_info[block][module] or {})
            # Update the options with the input file
            # Consistency is checked only up to first level! (i.e. subkeys may not match)
            # First deal with cases "no options" and "external function"
            try:
                input_info[block][module] = input_info[block][module] or {}
            except TypeError:
                log.error("Your input info is not well formatted at the '%s' block. "
                          "It must be a dictionary {'%s':{options}, ...}. ", block, block)
                raise HandledException
            if not hasattr(input_info[block][module], "get"):
                input_info[block][module] = {_external: input_info[block][module]}
            options_not_recognized = (set(input_info[block][module])
                                      .difference(set([_external, _p_renames]))
                                      .difference(set(full_info[block][module])))
            if options_not_recognized:
                alternatives = odict()
                available = set([_external, _p_renames]).union(full_info[block][module])
                while options_not_recognized:
                    option = options_not_recognized.pop()
                    try:
                        alternatives[option] = list(
                            zip(*(fuzzy_process.extractBests(
                                option, available, score_cutoff=50)[:3])))[0]
                    except IndexError:
                        alternatives[option] = []
                did_you_mean = ", ".join(
                    [("'%s' (did you mean %s?)" % (o,a) if a else "'%s'" % o)
                     for o,a in alternatives.items()])
                if default_module_info[block][module]:
                    # Internal module
                    log.error("'%s' does not recognize some options: %s. "
                              "To see the allowed options, check out the documentation of"
                              " this module.", module, did_you_mean)
                    raise HandledException
                else:
                    # External module
                    log.error("External %s '%s' does not recognize some options: %s. "
                              "Check the documentation for 'external %s'.",
                              block, module, did_you_mean, block)
                    raise HandledException
            full_info[block][module].update(input_info[block][module])
            # Store default parameters and priors of class, and save to combine later
            if block == _likelihood:
                params_info = default_module_info.get(_params, {})
                full_info[block][module].update({_params: list(params_info)})
                default_params_info[module] = params_info
                default_prior_info[module] = default_module_info.get(_prior, {})
    # Add priors info, after the necessary checks
    if _prior in input_info or any(default_prior_info.values()):
        full_info[_prior] = input_info.get(_prior, odict())
    for prior_info in default_prior_info.values():
        for name, prior in prior_info.items():
            if full_info[_prior].get(name, prior) != prior:
                log.error("Two different priors cannot have the same name: '%s'.", name)
                raise HandledException
            full_info[_prior][name] = prior
    # Add parameters info, after the necessary updates and checks
    defaults_merged = merge_default_params_info(default_params_info)
    full_info[_params] = merge_params_info(defaults_merged, input_info.get(_params, {}))
    # Add aliases for theory params (after merging!)
    if _theory in full_info:
        renames = list(full_info[_theory].values())[0].get(_p_renames)
        str_to_list = lambda x: ([x] if isinstance(x, string_types) else x)
        renames_flat = [set([k] + str_to_list(v)) for k, v in renames.items()]
        for p in full_info.get(_params, {}):
            # Probably could be made faster by inverting the renames dicts *just once*
            renames_pairs = [a for a in renames_flat if p in a]
            if renames_pairs:
                this_renames = reduce(
                    lambda x, y: x.union(y), [a for a in renames_flat if p in a])
                full_info[_params][p][_p_renames] = list(
                    set(this_renames).union(set(
                        str_to_list(full_info[_params][p].get(_p_renames, []))))
                        .difference(set([p])))
    # Rest of the options
    for k, v in input_info.items():
        if k not in full_info:
            full_info[k] = v
    return full_info


def merge_default_params_info(defaults):
    """
    Merges default parameters info for all likelihoods.
    Checks that multiple defined (=shared) parameters have equal info.
    """
    defaults_merged = odict()
    for lik, params in defaults.items():
        for p, info in params.items():
            # if already there, check consistency
            if p in defaults_merged:
                log.debug("Parameter '%s' multiply defined.", p)
                if info != defaults_merged[p]:
                    log.error("Parameter '%s' multiply defined, but inconsistent info: "
                              "For likelihood '%s' is '%r', but for some other likelihood"
                              " it was '%r'. Check your defaults!",
                              p, lik, info, defaults_merged[p])
                    raise HandledException
            defaults_merged[p] = info
    return defaults_merged


def merge_params_info(*params_infos):
    """
    Merges parameter infos, starting from the first one
    and updating with each additional one.
    Labels (for sampled and derived) and min/max
    (just for derived params) are inherited from defaults
    (but not if one of min/max is re-defined: in that case,
    to avoid surprises, the other one is set to None=+/-inf)
    """
    current_info = odict(
        [[p, expand_info_param(v)] for p, v in params_infos[0].items() or {}])
    for new_info in params_infos[1:]:
        if not new_info:
            continue
        for p, new_info_p in new_info.items():
            if p not in current_info:
                current_info[p] = odict()
            new_info_p = expand_info_param(new_info_p)
            current_info[p].update(deepcopy(new_info_p))
            # Account for incompatibilities: "prior" and ("value" or "derived"+bounds)
            incompatibilities = {_prior: [_p_value, _p_derived, "min", "max"],
                                 _p_value: [_prior, _p_ref, _p_proposal],
                                 _p_derived: [_prior, _p_drop, _p_ref, _p_proposal]}
            for f1, incomp in incompatibilities.items():
                if f1 in new_info_p:
                    for f2 in incomp:
                        current_info[p].pop(f2, None)
    # Re-sort, so that rightmost info takes precedence *also* in the sorting
    new_order = chain(*[list(params) for params in params_infos[::-1]])
    # The following removes duplicates maintaining order (keeps the first occurrence)
    new_order = list(odict.fromkeys(new_order))
    current_info = odict([[p, current_info[p]] for p in new_order])
    return current_info


def merge_info(*infos):
    """
    Merges information dictionaries. Rightmost arguments take precedence.
    """
    previous_info = deepcopy(infos[0])
    for new_info in infos[1:]:
        previous_params_info = deepcopy(previous_info.pop(_params, odict()) or odict())
        new_params_info = deepcopy(new_info).pop(_params, odict()) or odict()
        # NS: params have been popped, since they have their own merge function
        current_info = recursive_update(deepcopy(previous_info), new_info)
        current_info[_params] = merge_params_info(previous_params_info, new_params_info)
        previous_info = current_info
    return current_info


def is_equal_info(info1, info2, strict=True, print_not_log=False):
    """
    Compares two information dictionaries.

    Set ``strict=False`` (default: ``True``) to ignore options that would not affect
    the statistics of a posterior sample, including order of params/priors/likelihoods.
    """
    if print_not_log:
        myprint = print
    else:
        myprint = log.info
    myname = inspect.stack()[0][3]
    ignore = set([]) if strict else set(
        [_debug, _debug_file, _resume, _path_install])
    ignore_params = (set([]) if strict else set(
        [_p_label, _p_renames, _p_ref, _p_proposal, "min", "max"]))
    if set(info1).difference(ignore) != set(info2).difference(ignore):
        myprint(myname + ": different blocks or options: %r (old) vs %r (new)" % (
            set(info1).difference(ignore), set(info2).difference(ignore)))
        return False
    for block_name in info1:
        if block_name in ignore:
            continue
        block1, block2 = info1[block_name], info2[block_name]
        if not hasattr(block1, "keys"):
            if block1 != block2:
                myprint(myname + ": different option '%s'" % block_name)
                return False
        if block_name in [_sampler, _theory]:
            # Internal order does NOT matter
            if set(block1) != set(block2):
                myprint(myname + ": different [%s]" % block_name)
                return False
            # Anything to ignore?
            for k in block1:
                module_folder = get_folder(k, block_name, sep=".", absolute=False)
                try:
                    ignore_k = getattr(import_module(
                        module_folder, package=_package), "ignore_at_resume", {})
                except ImportError:
                    ignore_k = {}
                if block_name == _theory:
                    ignore_k.update({_input_params: None, _output_params: None})
                block1k, block2k = deepcopy(block1[k]), deepcopy(block2[k])
                if not strict:
                    for kignore in ignore_k:
                        try:
                            block1k.pop(kignore, None)
                            block2k.pop(kignore, None)
                        except:
                            pass
                if recursive_odict_to_dict(block1k) != recursive_odict_to_dict(block2k):
                    myprint(
                        myname + ": different content of [%s:%s]" % (block_name, k))
                    return False
        elif block_name in [_params, _likelihood, _prior]:
            # Internal order DOES matter, but just up to 1st level
            f = list if strict else set
            if f(block1) != f(block2):
                myprint(myname + ": different [%s] or different order of them: %r vs %r" % (
                    block_name, list(block1), list(block2)))
                return False
            for k in block1:
                block1k, block2k = deepcopy(block1[k]), deepcopy(block2[k])
                if block_name == _params:
                    # Unify notation
                    block1k = expand_info_param(block1k)
                    block2k = expand_info_param(block2k)
                    if not strict:
                        for kignore in ignore_params:
                            block1k.pop(kignore, None)
                            block2k.pop(kignore, None)
                        # Fixed params, it doesn't matter if they are saved as derived
                        for b in [block1k, block2k]:
                            if _p_value in b:
                                b.pop(_p_derived, None)
                if block_name == _likelihood and not strict:
                    for kignore in [_input_params, _output_params]:
                        block1k.pop(kignore, None)
                        block2k.pop(kignore, None)
                if (recursive_odict_to_dict(block1k) != recursive_odict_to_dict(block2k)):
                    myprint(myname + ": different content of [%s:%s]" % (
                        block_name, k))
                    return False
    return True



"""
.. module:: _planck_clik_prototype

:Synopsis: Definition of the clik-based likelihoods
:Author: Jesus Torrado
         (based on MontePython's version by Julien Lesgourgues and Benjamin Audren)

Family of Planck CMB likelihoods, based on the ``clik`` code.

The Planck 2015 likelihoods defined here are:

- ``planck_2015_lensing``
- ``planck_2015_lensing_cmblikes``
  (a native non-clik, more customizable version of the previous clik-wrapped one)
- ``planck_2015_lowl``
- ``planck_2015_lowTEB``
- ``planck_2015_plikHM_TT``
- ``planck_2015_plikHM_TTTEEE``
- ``planck_2015_plikHM_TTTEEE_unbinned``
- ``planck_2015_plikHM_TT_unbinned``

You can read a description of the different likelihoods in the
`Planck wiki <https://wiki.cosmos.esa.int/planckpla2015/index.php/CMB_spectrum_%26_Likelihood_Code>`_.

.. |br| raw:: html

   <br />

.. note::

   **If you use any of these likelihoods, please cite them as:**
   |br|
   N. Aghanim et al,
   `Planck 2015 results. XI. CMB power spectra, likelihoods, and robustness of parameters`
   `(arXiv:1507.02704) <https://arxiv.org/abs/1507.02704>`_

.. warning::

   ``planck_2015_lowTEB`` cannot be instantiated more than once.


Usage
-----

To use any of the Planck likelihoods, you simply need to mention them in the
``likelihood`` block, or add them using the :doc:`input generator <cosmo_basic_runs>`.

The corresponding nuisance parameters will be added automatically,
so you don't have to care about listing them in the ``params`` block.

The nuisance parameters and their default priors can be found in the ``defaults.yaml``
files in the folder corresponding to each likelihood. They are not reproduced here because
of their length.


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

.. note::

   By default, the ``gfortran`` compiler will be used, and the ``cfitsio`` library will be
   downloaded and compiled automatically.

   If the installation fails, make sure that the packages ``liblapack3`` and
   ``liblapack-dev`` are installed in the system (in Debian/Ubuntu, simply do
   ``sudo apt install liblapack3 liblapack-dev``).

   If you want to re-compile the Planck likelihood to your liking (e.g. with MKL), simply
   go into the chosen modules installation folder and re-run the ``python waf configure``
   and ``python waf install`` with the desired options.

However, if you wish to install it manually or have a previous installation already in
your system, simply take note of the path to the ``plc-2.0`` and ``plc_2.0`` folders and
mention it below each Planck likelihood as

.. code-block:: yaml

   likelihood:
     planck_2015_lowTEB:
       path: /path/to/planck_2015
     planck_2015_plikHM_TTTEEE:
       path: /path/to/planck_2015


Manual installation of Planck 2015 likelihoods
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. warning::

   For the time being (waiting for Planck 2018's data release), use instead the
   alternative 'clik' code at
   `<https://cdn.cosmologist.info/cosmobox/plc-2.1_py3.tar.bz2>`_, which is
   compatible with Python 3 and gcc `>5`.

Assuming you are installing all your likelihoods under ``/path/to/likelihoods``:

.. code:: bash

   $ cd /path/to/likelihoods
   $ mkdir planck_2015
   $ cd planck_2015
   $ wget https://pla.esac.esa.int/pla-sl/data-action?COSMOLOGY.COSMOLOGY_OID=1904
   $ tar xvjf data-action?COSMOLOGY.COSMOLOGY_OID=1904
   $ rm data-action?COSMOLOGY.COSMOLOGY_OID=1904
   $ cd plc-2.0
   $ python waf configure  # [options]

If the last step failed, try adding the option ``--install_all_deps``.
It it doesn't solve it, follow the instructions in the ``readme.md``
file in the ``plc-2.0`` folder.

If you have Intel's compiler and Math Kernel Library (MKL), you may want to also add the
option ``--lapack_mkl=${MKLROOT}`` in the last line to make use of it.

If ``python waf configure`` ended successfully run ``python waf install``
in the same folder. You do **not** need to run ``clik_profile.sh``, as advised.

Now, download the required likelihood files from the
`Planck Legacy Archive <https://pla.esac.esa.int/pla/#cosmology>`_ (Europe) or the
`NASA/IPAC Archive <https://irsa.ipac.caltech.edu/data/Planck/release_2/software/>`_ (US).

For instance, if you want to reproduce the baseline Planck 2015 results,
download the file ``COM_Likelihood_Data-baseline_R2.00.tar.gz``
from any of the two links above, and decompress it under the ``planck_2015`` folder
that you created above.

Finally, download and decompress in the ``planck_2015`` folder the last release at
`this repo <https://github.com/CobayaSampler/planck_supp_data_and_covmats/releases>`_.

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Global
import os
import sys
import numpy as np
import logging
import six

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import HandledException
from cobaya.conventions import _path_install, _likelihood
from cobaya.input import get_default_info
from cobaya.install import pip_install, download_file


class _planck_clik_prototype(Likelihood):

    def initialize(self):
        if self.path:
            has_clik = False
        else:
            try:
                import clik
                has_clik = True
            except ImportError:
                has_clik = False
        if not has_clik:
            if not self.path:
                if self.path_install:
                    self.path_clik = os.path.join(
                        self.path_install, "code", common_path)
                else:
                    self.log.error("No path given to the Planck likelihood. Set the "
                                   "likelihood property 'path' or the common property "
                                   "'%s'.", _path_install)
                    raise HandledException
            else:
                self.path_clik = self.path
            self.log.info("Importing clik from %s", self.path_clik)
            # test and import clik
            is_installed_clik(self.path_clik, log_and_fail=True, import_it=False)
            import clik
        # Loading the likelihood data
        if not os.path.isabs(self.clik_file):
            self.path_data = getattr(self, "path_data", os.path.join(
                self.path or self.path_install, "data", common_path))
            self.clik_file = os.path.join(self.path_data, self.clik_file)
        # for lensing, some routines change. Initializing a flag for easier
        # testing of this condition
        self.lensing = "lensing" in self.name
        try:
            self.clik = (
                clik.clik_lensing(self.clik_file) if self.lensing else clik.clik(self.clik_file))
        except clik.lkl.CError:
            self.log.error(
                "The .clik file was not found where specified in the 'clik_file' field "
                "of the settings of this likelihood. Maybe the 'path' given is not "
                "correct? The full path where the .clik file was searched for is '%s'",
                self.clik_file)
            raise HandledException
        self.expected_params = list(self.clik.extra_parameter_names)
        # py3 lensing bug
        if "b'A_planck'" in self.expected_params:
            self.expected_params[self.expected_params.index("b'A_planck'")] = "A_planck"
        # line added to deal with a bug in planck likelihood release:
        # A_planck called A_Planck in plik_lite
        if "plikHM_lite" in self.name:
            i = self.expected_params.index('A_Planck')
            self.expected_params[i] = 'A_planck'
            self.log.info("Corrected nuisance parameter name A_Planck to A_planck")
        # Check that the parameters are the right ones
        assert set(self.input_params) == set(self.expected_params), (
                "Likelihoods parameters do not coincide with the ones clik understands. Have: %s, Expected: %s"
                % (set(self.input_params), set(self.expected_params)))
        # Placeholder for vector passed to clik
        self.l_maxs = self.clik.get_lmax()
        length = (len(self.l_maxs) if self.lensing else len(self.clik.get_has_cl()))
        self.vector = np.zeros(np.sum(self.l_maxs) + length + len(self.expected_params))

    def add_theory(self):
        # State requisites to the theory code
        requested_cls = ["tt", "ee", "bb", "te", "tb", "eb"]
        if self.lensing:
            has_cl = [lmax != -1 for lmax in self.l_maxs]
            requested_cls = ["pp"] + requested_cls
        else:
            has_cl = self.clik.get_has_cl()
        self.requested_cls = [cl for cl, i in zip(requested_cls, has_cl) if int(i)]
        self.l_maxs_cls = [lmax for lmax, i in zip(self.l_maxs, has_cl) if int(i)]
        self.theory.needs(Cl=dict(zip(self.requested_cls, self.l_maxs_cls)))

    def logp(self, **params_values):
        # get Cl's from the theory code
        cl = self.theory.get_cl()
        # fill with Cl's
        self.vector[:-len(self.expected_params)] = np.concatenate(
            [(cl[spectrum][:1 + lmax] if spectrum not in ["tb", "eb"]
              else np.zeros(1 + lmax))
             for spectrum, lmax in zip(self.requested_cls, self.l_maxs_cls)])
        # fill with likelihood parameters
        self.vector[-len(self.expected_params):] = (
            [params_values[p] for p in self.expected_params])
        loglike = self.clik(self.vector)[0]
        # "zero" of clik
        if np.allclose(loglike, -1e30):
            loglike = -np.inf
        return loglike

    def close(self):
        del self.clik  # MANDATORY: forces deallocation of the Cython class
        # Actually, it does not work for low-l likelihoods, which is quite dangerous!


# Installation routines ##################################################################

# path to be shared by all Planck likelihoods
common_path = "planck_2015"

# To see full clik build output even if installs OK (e.g. to check warnings)
_clik_verbose = any(
    [(s in os.getenv('TRAVIS_COMMIT_MESSAGE', '')) for s in ["clik", "planck"]])
# Don't try again to install clik if it failed for a previous likelihood
_clik_install_failed = False


def is_installed_clik(path, log_and_fail=False, import_it=True):
    log = logging.getLogger("clik")
    if os.path.exists(path) and len(os.listdir(path)):
        clik_path = os.path.join(path, os.listdir(path)[0], 'lib/python/site-packages')
    else:
        clik_path = None
    if not clik_path or not os.path.exists(clik_path):
        if log_and_fail:
            log.error("The given folder does not exist: '%s'", clik_path or path)
            raise HandledException
        return False
    sys.path.insert(0, clik_path)
    try:
        if import_it:
            import clik
        return True
    except:
        print('Failed to import clik')
        if log_and_fail:
            log.error("Error importing click from: '%s'", clik_path)
            raise HandledException
        return False


def execute(command):
    from subprocess import Popen, PIPE, STDOUT
    if _clik_verbose:
        process = Popen(command, stdout=PIPE, stderr=STDOUT)
        out = []
        while True:
            nextline = process.stdout.readline()
            if nextline == b"" and process.poll() is not None:
                break
            if six.PY3:
                sys.stdout.buffer.write(nextline)
            else:
                sys.stdout.write(nextline)
            out.append(nextline)
            sys.stdout.flush()
        _, err = process.communicate()
        return b"finished successfully" in out[-1]
    else:
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        out, err = process.communicate()
        OK = b"finished successfully" in out.split(b"\n")[-2]
        if not OK:
            print(out.decode('utf-8'))
            print(err.decode('utf-8'))
        return OK


def install_clik(path, no_progress_bars=False):
    log = logging.getLogger("clik")
    log.info("Installing pre-requisites...")
    for req in ("cython", "astropy"):
        exit_status = pip_install(req)
        if exit_status:
            log.error("Failed installing '%s'.", req)
            raise HandledException
    log.info("Downloading...")
    click_url = 'https://cdn.cosmologist.info/cosmobox/plc-2.1_py3.tar.bz2'
    if not download_file(click_url, path, decompress=True,
                         no_progress_bars=no_progress_bars, logger=log):
        log.error("Not possible to download clik.")
        return False
    source_dir = os.path.join(path, os.listdir(path)[0])
    log.info('Installing from directory %s' % source_dir)
    if True:  # should be fixed
        log.info("Patching origin of cfitsio")
        cfitsio_filename = os.path.join(source_dir, "waf_tools", "cfitsio.py")
        with open(cfitsio_filename, "r") as cfitsio_file:
            lines = cfitsio_file.readlines()
            i_offending = next(i for i, l in enumerate(lines) if ".tar.gz" in l)
            lines[i_offending] = (
                "  atl.installsmthg_pre(ctx,"
                "'http://heasarc.gsfc.nasa.gov/FTP/software/fitsio/c/cfitsio3280.tar.gz',"
                "'cfitsio3280.tar.gz')\n")
        with open(cfitsio_filename, "w") as cfitsio_file:
            cfitsio_file.write("".join(lines))
    cwd = os.getcwd()
    try:
        os.chdir(source_dir)
        log.info("Configuring... (and maybe installing dependencies...)")
        if not execute([sys.executable, "waf", "configure", "--install_all_deps"]):
            log.error("Configuration failed!")
            return False

        log.info("Compiling...")
        if not execute([sys.executable, "waf", "install"]):
            log.error("Compilation failed!")
            return False
    finally:
        os.chdir(cwd)
    log.info("Finished!")
    return True


def get_product_id_and_clik_file(name):
    """Gets the PLA product info from the defaults file."""
    defaults = get_default_info(name, _likelihood)[_likelihood][name]
    return defaults.get("product_id"), defaults.get("clik_file")


def is_installed(**kwargs):
    result = True
    if kwargs["code"]:
        result &= is_installed_clik(os.path.realpath(
            os.path.join(kwargs["path"], "code", common_path)))
    if kwargs["data"]:
        _, filename = get_product_id_and_clik_file(kwargs["name"])
        result &= os.path.exists(os.path.realpath(
            os.path.join(kwargs["path"], "data", common_path, filename)))
        from cobaya.likelihoods.planck_2015_lensing_cmblikes import \
            is_installed as is_installed_supp
        result &= is_installed_supp(**kwargs)
    return result


def install(path=None, name=None, force=False, code=True, data=True,
            no_progress_bars=False):
    log = logging.getLogger(name)
    import platform
    if platform.system() == "Windows":
        log.error("Not compatible with Windows.")
        return False
    global _clik_install_failed
    if _clik_install_failed:
        log.info("Previous clik install failed, skipping")
        return False
    # Create common folders: all planck likelihoods share install folder for code and data
    paths = {}
    for s in ("code", "data"):
        if eval(s):
            paths[s] = os.path.realpath(os.path.join(path, s, common_path))
            if not os.path.exists(paths[s]):
                os.makedirs(paths[s])
    success = True
    # Install clik
    if code and (not is_installed_clik(paths["code"]) or force):
        log.info("Installing the clik code.")
        success *= install_clik(paths["code"], no_progress_bars=no_progress_bars)
        if not success:
            log.warning("clik code installation failed! "
                        "Try configuring+compiling by hand at " + paths["code"])
        _clik_install_failed = True
    if data:
        # 2nd test, in case the code wasn't there but the data is:
        if force or not is_installed(path=path, name=name, code=False, data=True):
            # Extract product_id
            product_id, _ = get_product_id_and_clik_file(name)
            # Download and decompress the particular likelihood
            log.info("Downloading likelihood data...")
            prefix = (r"https://pla.esac.esa.int/pla-sl/"
                      "data-action?COSMOLOGY.COSMOLOGY_OID=")
            if not download_file(prefix + product_id, paths["data"], decompress=True,
                                 logger=log, no_progress_bars=no_progress_bars):
                log.error("Not possible to download this likelihood.")
                success = False
            # Additional data and covmats
            from cobaya.likelihoods.planck_2015_lensing_cmblikes import \
                install as install_supp
            success *= install_supp(path=path, force=force, code=code, data=data,
                                    no_progress_bars=no_progress_bars)
    return success



"""
.. module:: samplers.fisher

:Synopsis: Fisher matrix estimator
:Author: ???
"""

# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function

# Global
import numpy as np
import logging

# Local
from cobaya.sampler import Sampler
from cobaya.log import HandledException
from cobaya.mpi import get_mpi_rank, get_mpi_size, get_mpi_comm
from cobaya.collection import Collection


class fisher(Sampler):
    def initialize(self):
        self.log.info("Initializing.")
        # Example for setting default options
        print("+++ For option '%s' got value '%s'. Add more options in fisher.yaml." % (
            "example_option", self.example_option))
        # Prepare the list of "samples" (posterior evaluations)
        self.collection = Collection(self.model, self.output)
        # Prepare vectors to store derivatives
        # ... up to you
        # Prepare the matrix
        self.fisher_matrix = np.eye(self.model.prior.d())

    def run(self):
        self.log.info("Starting Fisher matrix estimation.")
        # Getting info about parameters
        param_names = list(self.model.parameterization.sampled_params())
        print("Parameter names:", param_names)
        bounds = self.model.prior.bounds()
        print("Parameter bounds:", zip(param_names, bounds))
        # Get n samples from the prior, evaluate the likelihood,
        # and adding them to our collection of points
        #    (mind that self.model.prior.sample always returns an array, so,
        #     for a single sample do `sample = self.model.prior.sample()[0]`)
        initial_sample = self.model.prior.sample(4)
        for point in initial_sample:
            logpost, logpriors, loglikes, derived = self.model.logposterior(point)
            self.collection.add(
                point, logpost=logpost, logpriors=logpriors, loglikes=loglikes)
        print("--------------------")
        print("Our collection so far")
        print(self.collection)
        # Access one of the stored points and its values for parameters and logprobs
        i_point = 2  # the 3rd point!
        param = "x_01"
        print("For point #%d, " % i_point, "the value of parameter '%s' is %g" %
              (param, self.collection[i_point][param]))
        print("and the values of the logprior and loglikelihood are respectively %g, %g" %
              (-self.collection[i_point]["minuslogprior"],
               -0.5*self.collection[i_point]["chi2"]))
        # Accessing a column:
        print("All the values for parameter", param, "are\n", self.collection[param])
        # If you want to modify values for parameter/posteriors/etc, you have to COPY them
        # otherwise they'll be modified inside the table too!
        copy_of_minuslogpost = self.collection["minuslogpost"].values.copy()
        # Let me know if you need to know how to do anything else

    def close(self, *args):
        """
        Anything needed to be finished once convergence has been achieved (or not)
        """
        # Make sure that the last points are written
        self.collection._out_update()

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.
        """
        return {"evaluations": self.collection, "fisher_matrix": self.fisher_matrix}



import os
import logging

from cobaya.likelihoods._bao_prototype import _bao_prototype
from cobaya.install import download_github_release


class sdss_dr12_consensus_final(_bao_prototype):
    pass


# Installation routines ##################################################################

# name of the data and covmats repo/folder
sdss_data_name = "bao_data"
sdss_data_version = "v1.1"


def get_path(path):
    return os.path.realpath(os.path.join(path, "data", sdss_data_name))


def is_installed(**kwargs):
    return os.path.exists(os.path.realpath(
        os.path.join(kwargs["path"], "data", sdss_data_name)))


def install(path=None, force=False, code=False, data=True, no_progress_bars=False):
    if not data:
        return True
    log = logging.getLogger(__name__.split(".")[-1])
    log.info("Downloading BAO data...")
    return download_github_release(os.path.join(path, "data"), sdss_data_name,
                                   sdss_data_version, no_progress_bars=no_progress_bars)



"""
.. module:: collection

:Synopsis: Class to store the Montecarlo sample
:Author: Jesus Torrado

Class keeping track of the samples.

Basically, a wrapper around a `pandas.DataFrame`.

"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
import six

# Global
import os
from copy import deepcopy
import logging
import numpy as np
import pandas as pd
from getdist import MCSamples

# Local
from cobaya.conventions import _weight, _chi2, _minuslogpost, _minuslogprior
from cobaya.conventions import _separator
from cobaya.log import HandledException

# Suppress getdist output
from getdist import chains

chains.print_load_details = False

# Default chunk size for enlarging collections more efficiently
# If a factor is defined (as a fraction !=0), it is used instead
# (e.g. 0.10 to grow the number of rows by 10%)
enlargement_size = 100
enlargement_factor = None


# Make sure that we don't reach the empty part of the dataframe
def check_index(i, imax):
    if (i > 0 and i >= imax) or (i < 0 and -i > imax):
        raise IndexError("Trying to access a sample index larger than "
                         "the amount of samples (%d)!" % imax)
    if i < 0:
        return imax + i
    return i

# Notice that slices are never supposed to raise IndexError, but an empty list at worst!
def check_slice(ij, imax):
    newlims = {"start": ij.start, "stop": ij.stop}
    for limname, lim in newlims.items():
        if lim >= 0:
            newlims[limname] = min(imax, lim)
        else:
            newlims[limname] = imax + lim
    return slice(newlims["start"], newlims["stop"], ij.step)


class Collection(object):

    def __init__(self, model, output=None,
                 initial_size=enlargement_size, name=None, extension=None,
                 resuming=False, load=False, onload_skip=0, onload_thin=1):
        self.name = name
        self.log = logging.getLogger(
            "collection:" + name if name else self.__class__.__name__)
        self.sampled_params = list(model.parameterization.sampled_params())
        self.derived_params = list(model.parameterization.derived_params())
        self.minuslogprior_names = [
            _minuslogprior + _separator + piname for piname in list(model.prior)]
        self.chi2_names = [_chi2 + _separator + likname for likname in model.likelihood]
        # Create the dataframe structure
        columns = [_weight, _minuslogpost]
        columns += list(self.sampled_params)
        # Just in case: ignore derived names as likelihoods: would be duplicate cols
        columns += [p for p in self.derived_params if p not in self.chi2_names]
        columns += [_minuslogprior] + self.minuslogprior_names
        columns += [_chi2] + self.chi2_names
        # Create/load the main data frame and the tracking indices
        if output:
            self.file_name, self.driver = output.prepare_collection(
                name=self.name, extension=extension)
        else:
            self.driver = "dummy"
        if resuming or load:
            if output:
                try:
                    self._out_load(skip=onload_skip, thin=onload_thin)
                    if set(self.data.columns) != set(columns):
                        self.log.error(
                            "Unexpected column names!\nLoaded: %s\nShould be: %s",
                            list(self.data.columns), columns)
                        raise HandledException
                    self._n = self.data.shape[0]
                    self._n_last_out = self._n
                except IOError:
                    if resuming:
                        self.log.info(
                            "Could not find a chain to resume. "
                            "Maybe burn-in didn't finish. Creating new chain file!")
                        resuming = False
                    elif load:
                        raise
            else:
                self.log.error("No continuation possible if there is no output.")
                raise HandledException
        else:
            self._out_delete()
        if not resuming and not load:
            self.reset(columns=columns, index=range(initial_size))
            # TODO: the following 2 lines should go into the `reset` method.
            if output:
                self._n_last_out = 0

    def reset(self, columns=None, index=None):
        """Create/reset the DataFrame."""
        if columns is None:
            columns = self.data.columns
        if index is None:
            index = self.data.index
        self.data = pd.DataFrame(np.nan, columns=columns, index=index)
        self._n = 0

    def add(self,
            values, derived=None, weight=1, logpost=None, logpriors=None, loglikes=None):
        self._enlarge_if_needed()
        self.data.at[self._n, _weight] = weight
        if logpost is None:
            try:
                logpost = sum(logpriors) + sum(loglikes)
            except ValueError:
                self.log.error("If a log-posterior is not specified, you need to pass "
                               "a log-likelihood and a log-prior.")
                raise HandledException
        self.data.at[self._n, _minuslogpost] = -logpost
        if logpriors is not None:
            for name, value in zip(self.minuslogprior_names, logpriors):
                self.data.at[self._n, name] = -value
            self.data.at[self._n, _minuslogprior] = -sum(logpriors)
        if loglikes is not None:
            for name, value in zip(self.chi2_names, loglikes):
                self.data.at[self._n, name] = -2 * value
            self.data.at[self._n, _chi2] = -2 * sum(loglikes)
        if len(values) != len(self.sampled_params):
            self.log.error("Got %d values for the sampled parameters. Should be %d.",
                           len(values), len(self.sampled_params))
            raise HandledException
        for name, value in zip(self.sampled_params, values):
            self.data.at[self._n, name] = value
        if derived is not None:
            if len(derived) != len(self.derived_params):
                self.log.error("Got %d values for the dervied parameters. Should be %d.",
                               len(derived), len(self.derived_params))
                raise HandledException
            for name, value in zip(self.derived_params, derived):
                self.data.at[self._n, name] = value
        self._n += 1

    def _enlarge_if_needed(self):
        if self._n >= self.data.shape[0]:
            if enlargement_factor:
                enlarge_by = self.data.shape[0] * enlargement_factor
            else:
                enlarge_by = enlargement_size
            self.data = pd.concat([
                self.data, pd.DataFrame(np.nan, columns=self.data.columns,
                                        index=np.arange(self.n(), self.n() + enlarge_by))])

    def _append(self, collection):
        """
        Append another collection.
        Internal method: does not check for consistency!
        """
        self.data = pd.concat([self.data[:self.n()], collection.data], ignore_index=True)
        self._n = self.n() + collection.n()

    # Retrieve-like methods
    def n(self):
        return self._n

    def n_last_out(self):
        return self._n_last_out

    # Make the dataframe printable (but only the filled ones!)
    def __repr__(self):
        return self.data[:self.n()].__repr__()

    # Make the dataframe iterable over rows
    def __iter__(self):
        return self.data[:self.n()].iterrows()

    # Accessing the dataframe
    def __getitem__(self, *args):
        """
        This is a hack of the DataFrame __getitem__ in order to never go
        beyond the number of samples.

        NB: returns *views*, not *copies*, so careful when assigning and modifying
        the returned values.

        Errors are ValueError because this is not for internal use in this class,
        but an external interface to other classes and the user.
        """
        if len(args) > 1:
            raise ValueError("Use just one index/column, or use .loc[row, column]. "
                             "(Notice that slices in .loc *include* the last point.)")
        if isinstance(args[0], six.string_types):
            return self.data.iloc[:self._n, self.data.columns.get_loc(args[0])]
        elif hasattr(args[0], "__len__"):
            try:
                return self.data.iloc[:self._n,
                       [self.data.columns.get_loc(c) for c in args[0]]]
            except KeyError:
                raise ValueError("Some of the indices are not valid columns.")
        elif isinstance(args[0], six.integer_types):
            return self.data.iloc[check_index(args[0], self._n)]
        elif isinstance(args[0], slice):
            return self.data.iloc[check_slice(args[0], self._n)]
        else:
            raise ValueError("Index type not recognized: use column names or slices.")

    # Statistical computations
    def mean(self, first=None, last=None, derived=False):
        """
        Returns the (weighted) mean of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).
        """
        return np.average(
            self[list(self.sampled_params) +
                 (list(self.derived_params) if derived else [])]
            [first:last].T,
            weights=self[_weight][first:last], axis=-1)

    def cov(self, first=None, last=None, derived=False):
        """
        Returns the (weighted) mean of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).
        """
        weights = (lambda w: (
            {"fweights": w} if np.allclose(np.round(w), w) else {"aweights": w}))(
                self[_weight][first:last].values)
        return np.atleast_2d(np.cov(
            self[list(self.sampled_params) +
                 (list(self.derived_params) if derived else [])][first:last].T,
            **weights))

    def _sampled_to_getdist_mcsamples(self, first=None, last=None):
        """
        Basic interface with getdist -- internal use only!
        (For analysis and plotting use `getdist.mcsamples.loadCobayaSamples`.)
        """
        names = list(self.sampled_params)
        # No logging of warnings temporarily, so getdist won't complain unnecessarily
        logging.disable(logging.WARNING)
        mcsamples = MCSamples(
            samples=self.data[:self.n()][names].values[first:last],
            weights=self.data[:self.n()][_weight].values[first:last],
            loglikes=self.data[:self.n()][_minuslogpost].values[first:last], names=names)
        logging.disable(logging.NOTSET)
        return mcsamples

    # Copying and pickling
    def __deepcopy__(self, memo={}):
        new = (lambda cls: cls.__new__(cls))(self.__class__)
        new.__dict__  = {k: deepcopy(v) for k, v in self.__dict__.items() if k != "log"}
        return new

    def __getstate__(self):
        return deepcopy(self).__dict__

    # Saving and updating
    def _get_driver(self, method):
        return getattr(self, method + _separator + self.driver)

    # Load a pre-existing file
    def _out_load(self, **kwargs):
        self._get_driver("_load")(**kwargs)

    # Dump/update/delete collection
    def _out_dump(self):
        self._get_driver("_dump")()

    def _out_update(self):
        self._get_driver("_update")()

    def _out_delete(self):
        self._get_driver("_delete")()

    # txt driver
    def _load__txt(self, skip=0, thin=1):
        with open(self.file_name, "r") as inp:
            cols = [a.strip() for a in inp.readline().lstrip("#").split()]
            if 0 < skip < 1:
                # turn into #lines (need to know total line number)
                for n, line in enumerate(inp):
                    pass
                skip = int(skip * (n + 1))
                inp.seek(0)
            thin = int(thin)
            self.log.debug("Skipping %d rows and thinning with factor %d.", skip, thin)
            skiprows = lambda i: i < skip or i % thin
            self.data = pd.read_csv(
                inp, sep=" ", header=None, names=cols, comment="#", skipinitialspace=True,
                skiprows=skiprows)
        self.log.info("Loaded sample from '%s'", self.file_name)

    def _dump__txt(self):
        self._dump_slice__txt(0, self.n())

    def _update__txt(self):
        self._dump_slice__txt(self.n_last_out(), self.n())

    def _dump_slice__txt(self, n_min=None, n_max=None):
        if n_min is None or n_max is None:
            self.log.error("Needs to specify the limit n's to dump.")
            raise HandledException
        if self._n_last_out == n_max:
            return
        self._n_last_out = n_max
        n_float = 8
        do_header = not n_min
        with open(self.file_name, "a") as out:
            lines = self.data[n_min:n_max].to_string(
                header=do_header, index=False, na_rep="nan", justify="right",
                float_format=(lambda x: ("%%.%dg" % n_float) % x))
            # if header, add comment marker by hand (messes with align if auto)
            if do_header:
                lines = "#" + (lines[1:] if lines[0] == " " else lines)
            out.write(lines + "\n")

    def _delete__txt(self):
        try:
            os.remove(self.file_name)
        except OSError:
            pass

    # dummy driver
    def _dump__dummy(self):
        pass

    def _update__dummy(self):
        pass

    def _delete__dummy(self):
        pass


class OnePoint(Collection):
    """Wrapper of Collection to hold a single point, e.g. the current point of an MCMC."""

    def __init__(self, *args, **kwargs):
        kwargs["initial_size"] = 1
        Collection.__init__(self, *args, **kwargs)

    def __getitem__(self, columns):
        if isinstance(columns, six.string_types):
            return self.data.values[0, self.data.columns.get_loc(columns)]
        else:
            try:
                return self.data.values[0, [self.data.columns.get_loc(c) for c in columns]]
            except KeyError:
                raise ValueError("Some of the indices are not valid columns.")

    # Resets the counter, so the dataframe never fills up!
    def add(self, *args, **kwargs):
        Collection.add(self, *args, **kwargs)
        self._n = 0

    def increase_weight(self, increase):
        # For some reason, faster than `self.data[_weight] += increase`
        self.data.at[0, _weight] += increase

    def add_to_collection(self, collection):
        """Adds this point at the end of a given collection."""
        collection.add(
            self[self.sampled_params],
            derived=(self[self.derived_params] if self.derived_params else None),
            logpost=-self[_minuslogpost], weight=self[_weight],
            logpriors=-np.array(self[self.minuslogprior_names]),
            loglikes=-0.5 * np.array(self[self.chi2_names]))

    # Restore original __repr__ (here, there is only 1 sample)
    def __repr__(self):
        return self.data.__repr__()

    # Dumper changed: since we have made self._n=0, it would print nothing
    def _update__txt(self):
        self._dump_slice__txt(0, 1)



r"""
.. module:: one

:Synopsis: Unit likelihood: always outputs :math:`0 = \log(1)`.
:Author: Jesus Torrado

"""

# Python 2/3 compatibility
from __future__ import division

# Global
from random import random

# Local
from cobaya.likelihood import Likelihood


class one(Likelihood):
    """
    Likelihood that evaluates to 1.
    """

    def initialize(self):
        if self.noise:
            self.logp = self.logp_noise
        else:
            self.logp = self.logp_one

    def logp_one(self, **params_values):
        self.wait()
        return 0.

    def logp_noise(self, **params_values):
        self.wait()
        return self.noise * random()



try:
    from .gui import gui_script
except ImportError:
    # PySide not installed, but pass for now (will fail at GUI initialization)
    pass
from .autoselect_covmat import get_best_covmat
from .create_input import create_input
from .input_database import planck_base_model, cmb_precision, install_basic



from cobaya.likelihoods._planck_clik_prototype import _planck_clik_prototype


class planck_2015_lowTEB(_planck_clik_prototype):
    pass


from cobaya.likelihoods._planck_clik_prototype import install as install_common
from cobaya.likelihoods._planck_clik_prototype import is_installed as is_installed_common


def is_installed(**kwargs):
    kwargs["name"] = __name__.split(".")[-1]
    return is_installed_common(**kwargs)


def install(**kwargs):
    kwargs["name"] = __name__.split(".")[-1]
    return install_common(**kwargs)



"""
.. module:: post

:Synopsis: Post-processing functions
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division

# Global
import numpy as np
import logging
from collections import OrderedDict as odict
from numbers import Number
from copy import deepcopy

# Local
from cobaya.input import load_input
from cobaya.parameterization import Parameterization, expand_info_param
from cobaya.parameterization import is_fixed_param, is_sampled_param, is_derived_param
from cobaya.conventions import _prior_1d_name, _debug, _debug_file, _output_prefix, _post
from cobaya.conventions import _params, _prior, _likelihood, _theory, _p_drop, _weight
from cobaya.conventions import _chi2, _separator, _minuslogpost, _force, _p_value, _p_drop
from cobaya.conventions import _minuslogprior, _path_install, _input_params
from cobaya.collection import Collection
from cobaya.log import logger_setup, HandledException
from cobaya.input import get_full_info
from cobaya.output import get_Output as Output
from cobaya.prior import Prior
from cobaya.likelihood import LikelihoodCollection as Likelihood
from cobaya.mpi import get_mpi_rank
from cobaya.tools import progress_bar, recursive_update


# Dummy classes for loading chains for post processing

class DummyModel(object):

    def __init__(self, info_params, info_likelihood, info_prior=None, info_theory=None):

        self.parameterization = Parameterization(info_params, ignore_unused_sampled=True)
        self.prior = [_prior_1d_name] + list(info_prior or [])
        self.likelihood = list(info_likelihood)


def post(info, sample=None):
    logger_setup(info.get(_debug), info.get(_debug_file))
    log = logging.getLogger(__name__.split(".")[-1])
    try:
        info_post = info[_post]
    except KeyError:
        log.error("No 'post' block given. Nothing to do!")
        raise HandledException
    if get_mpi_rank():
        log.warning(
            "Post-processing is not yet MPI-able. Doing nothing for rank > 1 processes.")
        return
    # 1. Load existing sample
    output_in = Output(output_prefix=info.get(_output_prefix), resume=True)
    info_in = load_input(output_in.file_full) if output_in else deepcopy(info)
    dummy_model_in = DummyModel(info_in[_params], info_in[_likelihood],
                                info_in.get(_prior, None), info_in.get(_theory, None))
    if output_in:
        i = 0
        while True:
            try:
                collection = Collection(
                    dummy_model_in, output_in, name="%d" % (1 + i),
                    load=True, onload_skip=info_post.get("skip", 0),
                    onload_thin=info_post.get("thin", 1))
                if i == 0:
                    collection_in = collection
                else:
                    collection_in._append(collection)
                i += 1
            except IOError:
                break
    elif sample:
        if isinstance(sample, Collection):
            sample = [sample]
        collection_in = deepcopy(sample[0])
        for s in sample[1:]:
            try:
                collection_in._append(s)
            except:
                log.error("Failed to load some of the input samples.")
                raise HandledException
        i = len(sample)
    else:
        log.error("Not output from where to load from or input collections given.")
        raise HandledException
    log.info("Loaded %d chain%s. Will process %d samples.",
             i, "s" if i - 1 else "", collection_in.n())
    if collection_in.n() <= 1:
        log.error("Not enough samples for post-processing. Try using a larger sample, "
                  "or skipping or thinning less.")
        raise HandledException
    # 2. Compare old and new info: determine what to do
    add = info_post.get("add", {})
    remove = info_post.get("remove", {})
    # Add a dummy 'one' likelihood, to absorb unused parameters
    if not add.get(_likelihood):
        add[_likelihood] = odict()
    add[_likelihood].update({"one": None})
    # Expand the "add" info
    add = get_full_info(add)
    # 2.1 Adding/removing derived parameters and changes in priors of sampled parameters
    out = {_params: deepcopy(info_in[_params])}
    for p in remove.get(_params, {}):
        pinfo = info_in[_params].get(p)
        if pinfo is None or not is_derived_param(pinfo):
            log.error(
                "You tried to remove parameter '%s', which is not a derived paramter. "
                "Only derived parameters can be removed during post-processing.", p)
            raise HandledException
        out[_params].pop(p)
    mlprior_names_add = []
    for p, pinfo in add.get(_params, {}).items():
        pinfo_in = info_in[_params].get(p)
        if is_sampled_param(pinfo):
            if not is_sampled_param(pinfo_in):
                # No added sampled parameters (de-marginalisation not implemented)
                if pinfo_in is None:
                    log.error("You added a new sampled parameter %r (maybe accidentaly "
                              "by adding a new likelihood that depends on it). "
                              "Adding new sampled parameters is not possible. Try fixing "
                              "it to some value.", p)
                    raise HandledException
                else:
                    log.error(
                        "You tried to change the prior of parameter '%s', "
                        "but it was not a sampled parameter. "
                        "To change that prior, you need to define as an external one.", p)
                    raise HandledException
            if mlprior_names_add[:1] != _prior_1d_name:
                mlprior_names_add = (
                    [_minuslogprior + _separator + _prior_1d_name] + mlprior_names_add)
        elif is_derived_param(pinfo):
            if p in out[_params]:
                log.error("You tried to add derived parameter '%s', which is already "
                          "present. To force its recomputation, 'remove' it too.", p)
                raise HandledException
        elif is_fixed_param(pinfo):
            # Only one possibility left "fixed" parameter that was not present before:
            # input of new likelihood, or just an argument for dynamical derived (dropped)
            if ((p in info_in[_params] and
                 pinfo[_p_value] != (pinfo_in or {}).get(_p_value, None))):
                log.error(
                    "You tried to add a fixed parameter '%s: %r' that was already present"
                    " but had a different value or was not fixed. This is not allowed. "
                    "The old info of the parameter was '%s: %r'",
                    p, dict(pinfo), p, dict(pinfo_in))
                raise HandledException
        else:
            log.error("This should not happen. Contact the developers.")
            raise HandledException
        out[_params][p] = pinfo
    # For the likelihood only, turn the rest of *derived* parameters into constants,
    # so that the likelihoods do not try to compute them)
    # But be careful to exclude *input* params that have a "derived: True" value
    # (which in "full info" turns into "derived: 'lambda [x]: [x]'")
    out_params_like = deepcopy(out[_params])
    for p, pinfo in out_params_like.items():
        if ((is_derived_param(pinfo) and not(_p_value in pinfo)
             and p not in add.get(_params, {}))):
            out_params_like[p] = {_p_value: np.nan, _p_drop: True}
    parameterization_like = Parameterization(out_params_like, ignore_unused_sampled=True)
    # 2.2 Manage adding/removing priors and likelihoods
    warn_remove = False
    for level in [_prior, _likelihood]:
        out[level] = getattr(dummy_model_in, level)
        if level == _prior:
            out[level].remove(_prior_1d_name)
        for pdf in info_post.get("remove", {}).get(level, []) or []:
            try:
                out[level].remove(pdf)
                warn_remove = True
            except ValueError:
                log.error("Trying to remove %s '%s', but it is not present. "
                          "Existing ones: %r", level, pdf, out[level])
                raise HandledException
    if warn_remove:
        log.warning("You are removing a prior or likelihood pdf. "
                    "Notice that if the resulting posterior is much wider "
                    "than the original one, or displaced enough, "
                    "it is probably safer to explore it directly.")
    if _prior in add:
        mlprior_names_add += [_minuslogprior + _separator + name for name in add[_prior]]
        out[_prior] += list(add[_prior])
    prior_recompute_1d = (
        mlprior_names_add[:1] == [_minuslogprior + _separator + _prior_1d_name])
    # Don't initialise the theory code if not adding/recomputing theory,
    # theory-derived params or likelihoods
    recompute_theory = info_in.get(_theory) and not (
        list(add[_likelihood]) == ["one"] and
        not any([is_derived_param(pinfo) for pinfo in add.get(_params, {}).values()]))
    if recompute_theory:
        # Inherit from the original chain (needs input|output_params, renames, etc
        theory = list(info_in[_theory].keys())[0]
        info_theory_out = odict([
            [theory, recursive_update(deepcopy(info_in[_theory][theory]),
                                      add.get(_theory, {theory: {}})[theory])]])
    else:
        info_theory_out = None
    chi2_names_add = [_chi2 + _separator + name for name in add[_likelihood]
                      if name is not "one"]
    out[_likelihood] += [l for l in add[_likelihood] if l is not "one"]
    if recompute_theory:
        log.warn("You are recomputing the theory, but in the current version this does "
                 "not force recomputation of any likelihood or derived parameter, "
                 "unless explicitly removed+added.")
    for level in [_prior, _likelihood]:
        for i, x_i in enumerate(out[level]):
            if x_i in list(out[level])[i+1:]:
                log.error("You have added %s '%s', which was already present. If you "
                          "want to force its recomputation, you must also 'remove' it.",
                          level, x_i)
                raise HandledException
    # 3. Create output collection
    if "suffix" not in info_post:
        log.error("You need to provide a 'suffix' for your chains.")
        raise HandledException
    # Use default prefix if it exists. If it does not, produce no output by default.
    # {post: {output: None}} suppresses output, and if it's a string, updates it.
    out_prefix = info_post.get(_output_prefix, info.get(_output_prefix))
    if out_prefix not in [None, False]:
        out_prefix += "_" + _post + "_" + info_post["suffix"]
    output_out = Output(output_prefix=out_prefix, force_output=info.get(_force))
    info_out = deepcopy(info)
    info_out[_post] = info_post
    # Updated with input info and extended (full) add info
    info_out.update(info_in)
    info_out[_post]["add"] = add
    dummy_model_out = DummyModel(
        out[_params], out[_likelihood], info_prior=out[_prior])
    if recompute_theory:
        theory = list(info_theory_out.keys())[0]
        if _input_params not in info_theory_out[theory]:
            log.error(
                "You appear to be post-processing a chain generated with an older "
                "version of Cobaya. For post-processing to work, please edit the "
                "'[root]__full.info' file of the original chain to add, inside the "
                "theory code block, the list of its input parameters. E.g.\n----\n"
                "theory:\n  %s:\n    input_params: [param1, param2, ...]\n"
                "----\nIf you get strange errors later, it is likely that you did not "
                "specify the correct set of theory parameters.\n"
                "The full set of input parameters are %s.",
                theory, list(dummy_model_out.parameterization.input_params()))
            raise HandledException
    prior_add = Prior(dummy_model_out.parameterization, add.get(_prior))
    likelihood_add = Likelihood(
        add[_likelihood], parameterization_like,
        info_theory=info_theory_out, modules=info.get(_path_install))
    # Remove auxiliary "one" before dumping -- 'add' *is* info_out[_post]["add"]
    add[_likelihood].pop("one")
    if likelihood_add.theory:
        # Make sure that theory.needs is called at least once, for adjustments
        likelihood_add.theory.needs()
    collection_out = Collection(dummy_model_out, output_out, name="1")
    output_out.dump_info({}, info_out)
    # 4. Main loop!
    log.info("Running post-processing...")
    last_percent = 0
    for i, point in enumerate(collection_in.data.itertuples()):
        log.debug("Point: %r", point)
        sampled = [getattr(point, param) for param in
                   dummy_model_in.parameterization.sampled_params()]
        derived = odict(
            [[param, getattr(point, param, None)]
             for param in dummy_model_out.parameterization.derived_params()])
        inputs = odict([
            [param, getattr(
                point, param,
                dummy_model_in.parameterization.constant_params().get(
                    param, dummy_model_out.parameterization.constant_params().get(
                        param, None)))]
            for param in dummy_model_out.parameterization.input_params()])
        # Solve inputs that depend on a function and were not saved
        # (we don't use the Parameterization_to_input method in case there are references
        #  to functions that cannot be loaded at the moment)
        for p, value in inputs.items():
            if value is None:
                func = dummy_model_out.parameterization._input_funcs[p]
                args = dummy_model_out.parameterization._input_args[p]
                inputs[p] = func(*[getattr(point, arg) for arg in args])
        # Add/remove priors
        priors_add = prior_add.logps(sampled)
        if not prior_recompute_1d:
            priors_add = priors_add[1:]
        logpriors_add = odict(zip(mlprior_names_add, priors_add))
        logpriors_new = [logpriors_add.get(name, - getattr(point, name, 0))
                         for name in collection_out.minuslogprior_names]
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug(
                "New set of priors: %r", dict(zip(dummy_model_out.prior, logpriors_new)))
        if -np.inf in logpriors_new:
            continue
        # Add/remove likelihoods
        output_like = []
        if likelihood_add:
            # Notice "one" (last in likelihood_add) is ignored: not in chi2_names
            loglikes_add = odict(
                zip(chi2_names_add, likelihood_add.logps(inputs, _derived=output_like)))
            output_like = dict(zip(likelihood_add.output_params, output_like))
        else:
            loglikes_add = dict()
        loglikes_new = [loglikes_add.get(name, -0.5 * getattr(point, name, 0))
                        for name in collection_out.chi2_names]
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug(
                "New set of likelihoods: %r",
                dict(zip(dummy_model_out.likelihood, loglikes_new)))
            if output_like:
                log.debug("New set of likelihood-derived parameters: %r", output_like)
        if -np.inf in loglikes_new:
            continue
        # Add/remove derived parameters and change priors of sampled parameters
        for p in add[_params]:
            if p in dummy_model_out.parameterization._directly_output:
                derived[p] = output_like[p]
            elif p in dummy_model_out.parameterization._derived_funcs:
                func = dummy_model_out.parameterization._derived_funcs[p]
                args = dummy_model_out.parameterization._derived_args[p]
                derived[p] = func(
                    *[getattr(point, arg, output_like.get(arg, None)) for arg in args])
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("New derived parameters: %r",
                      dict([[p, derived[p]]
                            for p in dummy_model_out.parameterization.derived_params()
                            if p in add[_params]]))
        # Save to the collection (keep old weight for now)
        collection_out.add(
            sampled, derived=derived.values(), weight=getattr(point, _weight),
            logpriors=logpriors_new, loglikes=loglikes_new)
        # Display progress
        percent = np.round(i / collection_in.n() * 100)
        if percent != last_percent and not percent % 5:
            last_percent = percent
            progress_bar(log, percent, " (%d/%d)" % (i, collection_in.n()))
    if not collection_out.data.last_valid_index():
        log.error("No elements in the final sample. Possible causes: "
                  "added a prior or likelihood valued zero over the full sampled domain, "
                  "or the computation of the theory failed everywhere, etc.")
        raise HandledException
    # Reweight -- account for large dynamic range!
    #   Prefer to rescale +inf to finite, and ignore final points with -inf.
    #   Remove -inf's (0-weight), and correct indices
    difflogmax = max(collection_in[_minuslogpost] - collection_out[_minuslogpost])
    collection_out.data[_weight] *= np.exp(
        collection_in[_minuslogpost] - collection_out[_minuslogpost] - difflogmax)
    collection_out.data = (
        collection_out.data[collection_out.data.weight > 0].reset_index(drop=True))
    collection_out._n = collection_out.data.last_valid_index() + 1
    # Write!
    collection_out._out_update()
    log.info("Finished! Final number of samples: %d", collection_out.n())
    return info_out, {"sample": collection_out}
